import numpy as np
from libensemble.message_numbers import EVAL_GEN_TAG
from libensemble.tools.alloc_support import AllocSupport, InsufficientFreeResources


def persistent_aposmm_alloc(W, H, sim_specs, gen_specs, alloc_specs, persis_info, libE_info):
    """
    This allocation function will give simulation work if possible, but
    otherwise start a persistent APOSMM generator.  If all points requested by
    the persistent generator have been returned from the simulation evaluation,
    then this information is given back to the persistent generator.

    This function assumes that one persistent APOSMM will be started and never
    stopped (until some exit_criterion is satisfied).

    .. seealso::
        `test_persistent_aposmm_with_grad.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_persistent_aposmm_with_grad.py>`_ # noqa
    """

    if libE_info['sim_max_given'] or not libE_info['any_idle_workers']:
        return {}, persis_info

    # Initialize alloc_specs['user'] as user.
    user = alloc_specs.get('user', {})
    sched_opts = user.get('scheduler_opts', {})

    init_sample_size = gen_specs['user']['initial_sample_size']
    manage_resources = 'resource_sets' in H.dtype.names or libE_info['use_resource_sets']
    support = AllocSupport(W, manage_resources, persis_info, sched_opts)
    gen_count = support.count_persis_gens()
    Work = {}

    if persis_info.get('first_call', True):
        assert support.all_given(H), "Initial points in H have never been given."
        assert support.all_returned(H), "Initial points in H have never been returned."
        assert support.all_given_back(H), "Initial points in H have never been given back to gen."

        persis_info['samples_in_H0'] = sum(H['local_pt'] == 0)
        persis_info['next_to_give'] = len(H)  #
        persis_info['first_call'] = False
    elif gen_count == 0:
        # The one persistent gen is done. Exiting
        return Work, persis_info, 1

    # If any persistent worker's calculated values have returned, give them back.
    for wid in support.avail_worker_ids(persistent=EVAL_GEN_TAG):
        if (persis_info.get('sample_done') or sum(H['returned']) >= init_sample_size + persis_info['samples_in_H0']):
            # Don't return if the initial sample is not complete
            persis_info['sample_done'] = True

            returned_but_not_given = np.logical_and(H['returned'], ~H['given_back'])
            if np.any(returned_but_not_given):
                point_ids = np.where(returned_but_not_given)[0]
                Work[wid] = support.gen_work(wid, gen_specs['persis_in'], point_ids, persis_info.get(wid),
                                             persistent=True)
                returned_but_not_given[point_ids] = False

    for wid in support.avail_worker_ids(persistent=False):
        # Skip any cancelled points
        while persis_info['next_to_give'] < len(H) and H[persis_info['next_to_give']]['cancel_requested']:
            persis_info['next_to_give'] += 1

        if persis_info['next_to_give'] < len(H):
            # perform sim evaluations (if they exist in History).
            try:
                Work[wid] = support.sim_work(wid, H, sim_specs['in'], persis_info['next_to_give'], persis_info.get(wid))
            except InsufficientFreeResources:
                break
            persis_info['next_to_give'] += 1

        elif persis_info.get('gen_started') is None:
            # Finally, call a persistent generator as there is nothing else to do.
            persis_info.get(wid)['nworkers'] = len(W)
            try:
                Work[wid] = support.gen_work(wid, gen_specs.get('in', []), range(len(H)), persis_info.get(wid),
                                             persistent=True)
            except InsufficientFreeResources:
                break
            persis_info['gen_started'] = True  # Must set after - in case break on resources

    return Work, persis_info



"""
libEnsemble.

Library to coordinate the concurrent evaluation of dynamic ensembles of calculations.
"""

from .version import __version__
__author__ = 'Jeffrey Larson, Stephen Hudson, Stefan M. Wild, David Bindel and John-Luke Navarro'
__credits__ = 'Argonne National Laboratory'

from libensemble import logger

from .libE import libE, comms_abort

try:
    from .api import Ensemble
except ModuleNotFoundError:
    pass



"""
Runs libEnsemble with a non-persistent generator performing uniform random
sampling.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_uniform_sampling.py
   python3 test_uniform_sampling.py --nworkers 3 --comms local
   python3 test_uniform_sampling.py --nworkers 3 --comms tcp

The number of concurrent evaluations of the objective function will be 4-1=3.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 2 4

import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.six_hump_camel import six_hump_camel
from libensemble.gen_funcs.sampling import uniform_random_sample
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams
from libensemble.tests.regression_tests.support import six_hump_camel_minima as minima

nworkers, is_manager, libE_specs, _ = parse_args()
libE_specs['save_every_k_sims'] = 400
libE_specs['save_every_k_gens'] = 300

sim_specs = {
    'sim_f': six_hump_camel,  # Function whose output is being minimized
    'in': ['x'],  # Keys to be given to sim_f
    'out': [('f', float)],  # Name of the outputs from sim_f
}
# end_sim_specs_rst_tag

gen_specs = {
    'gen_f': uniform_random_sample,  # Function generating sim_f input
    'out': [('x', float, (2,))],  # Tell libE gen_f output, type, size
    'user': {
        'gen_batch_size': 500,  # Used by this specific gen_f
        'lb': np.array([-3, -2]),  # Used by this specific gen_f
        'ub': np.array([3, 2]),  # Used by this specific gen_f
    },
}
# end_gen_specs_rst_tag

persis_info = add_unique_random_streams({}, nworkers + 1)

exit_criteria = {'gen_max': 501, 'elapsed_wallclock_time': 300}

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, libE_specs=libE_specs)

if is_manager:
    assert flag == 0

    tol = 0.1
    for m in minima:
        assert np.min(np.sum((H['x'] - m) ** 2, 1)) < tol

    print("\nlibEnsemble found the 6 minima within a tolerance " + str(tol))
    save_libE_output(H, persis_info, __file__, nworkers)



"""
This module contains methods for APOSMM to interface with various local
optimization routines.
"""
__all__ = ['LocalOptInterfacer', 'run_local_nlopt', 'run_local_tao',
           'run_local_dfols', 'run_local_scipy_opt', 'run_external_localopt']

import psutil
from libensemble.tools.tools import osx_set_mp_method
import numpy as np
from libensemble.message_numbers import STOP_TAG, EVAL_GEN_TAG  # Only used to simulate receiving from manager
from multiprocessing import Event, Process, Queue
import libensemble.gen_funcs

optimizer_list = ['petsc', 'nlopt', 'dfols', 'scipy', 'external']
optimizers = libensemble.gen_funcs.rc.aposmm_optimizers

# Resolves multiprocessing issues with Python 3.8+ on macOS
osx_set_mp_method()

if optimizers is None:
    from petsc4py import PETSc
    import nlopt
    import dfols
    from scipy import optimize as sp_opt
else:
    if not isinstance(optimizers, list):
        optimizers = [optimizers]
    unrec = set(optimizers) - set(optimizer_list)
    if unrec:
        print('APOSMM Warning: unrecognized optimizers {}'.format(unrec))

    if 'petsc' in optimizers:
        from petsc4py import PETSc
    if 'nlopt' in optimizers:
        import nlopt
    if 'dfols' in optimizers:
        import dfols
    if 'scipy' in optimizers:
        from scipy import optimize as sp_opt
    if 'external' in optimizers:
        pass


class APOSMMException(Exception):
    "Raised for any exception in APOSMM"


class ConvergedMsg(object):
    """
    Message communicated when a local optimization is converged.
    """
    def __init__(self, x, opt_flag):
        self.x = x
        self.opt_flag = opt_flag


class ErrorMsg(object):
    """
    Message communicated when a local optimization has an exception.
    """
    def __init__(self, x):
        self.x = x


class LocalOptInterfacer(object):
    """
    This class defines the APOSMM interface to various local optimization routines.

    Currently supported routines are

    - NLopt routines ['LN_SBPLX', 'LN_BOBYQA', 'LN_COBYLA', 'LN_NEWUOA', 'LN_NELDERMEAD', 'LD_MMA']
    - PETSc/TAO routines ['pounders', 'blmvm', 'nm']
    - SciPy routines ['scipy_Nelder-Mead', 'scipy_COBYLA', 'scipy_BFGS']
    - DFOLS ['dfols']
    - External local optimizer ['external_localopt'] (which use files to pass/receive x/f values)
    """

    def __init__(self, user_specs, x0, f0, grad0=None):
        """
        :param x0: A numpy array of the initial guess solution. This guess
            should be scaled to a unit cube.
        :param f0: A numpy array of the initial function value.


        .. warning:: In order to have correct functioning of the local
            optimization child processes. ~self.iterate~ should be called
            immediately after creating the class.

        """

        self.parent_can_read = Event()
        self.comm_queue = Queue()
        self.child_can_read = Event()

        self.x0 = x0.copy()
        self.f0 = f0.copy()
        if grad0 is not None:
            self.grad0 = grad0.copy()
        else:
            self.grad0 = None

        # Setting the local optimization method
        if user_specs['localopt_method'] in ['LN_SBPLX', 'LN_BOBYQA', 'LN_COBYLA',
                                             'LN_NEWUOA', 'LN_NELDERMEAD', 'LD_MMA']:
            run_local_opt = run_local_nlopt
        elif user_specs['localopt_method'] in ['pounders', 'blmvm', 'nm']:
            run_local_opt = run_local_tao
        elif user_specs['localopt_method'] in ['scipy_Nelder-Mead', 'scipy_COBYLA', 'scipy_BFGS']:
            run_local_opt = run_local_scipy_opt
        elif user_specs['localopt_method'] in ['dfols']:
            run_local_opt = run_local_dfols
        elif user_specs['localopt_method'] in ['external_localopt']:
            run_local_opt = run_external_localopt

        self.parent_can_read.clear()
        self.process = Process(target=opt_runner, args=(run_local_opt, user_specs,
                               self.comm_queue, x0, f0, self.child_can_read,
                               self.parent_can_read))

        self.process.start()
        self.is_running = True
        self.parent_can_read.wait()
        x_new = self.comm_queue.get()
        if isinstance(x_new, ErrorMsg):
            raise APOSMMException(x_new.x)

        assert np.allclose(x_new, x0, rtol=1e-15, atol=1e-15), \
            "The first point requested by this run does not match the starting point. Exiting"

    def iterate(self, data):
        """
        Returns an instance of either :class:`numpy.ndarray` corresponding to the next
        iterative guess or :class:`ConvergedMsg` when the solver has completed its run.

        :param x_on_cube: A numpy array of the point being evaluated (for a handshake)
        :param f: A numpy array of the function evaluation.
        :param grad: A numpy array of the function's gradient.
        :param fvec: A numpy array of the function's component values.
        """

        self.parent_can_read.clear()

        if 'grad' in data.dtype.names:
            self.comm_queue.put((data['x_on_cube'], data['f'], data['grad']))
        elif 'fvec' in data.dtype.names:
            self.comm_queue.put((data['x_on_cube'], data['fvec']))
        else:
            self.comm_queue.put((data['x_on_cube'], data['f'], ))

        self.child_can_read.set()
        self.parent_can_read.wait()

        x_new = self.comm_queue.get()
        if isinstance(x_new, ErrorMsg):
            raise APOSMMException(x_new.x)
        elif isinstance(x_new, ConvergedMsg):
            self.close()
        else:
            x_new = np.atleast_2d(x_new)

        return x_new

    def destroy(self):
        """Recursively kill any optimizer processes still running"""
        if self.process.is_alive():
            process = psutil.Process(self.process.pid)
            for child in process.children(recursive=True):
                child.kill()
            process.kill()
        self.close()

    def close(self):
        """Join process and close queue"""
        self.process.join()
        self.comm_queue.close()
        self.comm_queue.join_thread()
        self.is_running = False


def run_local_nlopt(user_specs, comm_queue, x0, f0, child_can_read, parent_can_read):
    """
    Runs an NLopt local optimization run starting at ``x0``, governed by the
    parameters in ``user_specs``.
    """
    # print('[Child]: Started local opt at {}.'.format(x0), flush=True)
    n = len(user_specs['ub'])

    opt = nlopt.opt(getattr(nlopt, user_specs['localopt_method']), n)

    lb = np.zeros(n)
    ub = np.ones(n)

    if not user_specs.get('periodic'):
        opt.set_lower_bounds(lb)
        opt.set_upper_bounds(ub)

    # Care must be taken here because a too-large initial step causes nlopt to move the starting point!
    dist_to_bound = min(min(ub-x0), min(x0-lb))
    assert dist_to_bound > np.finfo(np.float64).eps, "The distance to the boundary is too small for NLopt to handle"

    if 'dist_to_bound_multiple' in user_specs:
        opt.set_initial_step(dist_to_bound*user_specs['dist_to_bound_multiple'])
    else:
        opt.set_initial_step(dist_to_bound)

    run_max_eval = user_specs.get('run_max_eval', 1000*n)
    opt.set_maxeval(run_max_eval)

    opt.set_min_objective(lambda x, grad: nlopt_callback_fun(x, grad,
                          comm_queue, child_can_read, parent_can_read,
                          user_specs))

    if 'xtol_rel' in user_specs:
        opt.set_xtol_rel(user_specs['xtol_rel'])
    if 'ftol_rel' in user_specs:
        opt.set_ftol_rel(user_specs['ftol_rel'])
    if 'xtol_abs' in user_specs:
        opt.set_xtol_abs(user_specs['xtol_abs'])
    if 'ftol_abs' in user_specs:
        opt.set_ftol_abs(user_specs['ftol_abs'])

    # FIXME: Do we need to do something of the final 'x_opt'?
    # print('[Child]: Started my optimization', flush=True)
    x_opt = opt.optimize(x0)
    return_val = opt.last_optimize_result()

    if return_val >= 1 and return_val <= 4:
        # These return values correspond to an optimium being identified
        # https://nlopt.readthedocs.io/en/latest/NLopt_Reference/#return-values
        opt_flag = 1
    elif return_val >= 5:
        print("[APOSMM] The run started from " + str(x0) + " reached its maximum number "
              "of function evaluations: " + str(run_max_eval) + ". No point from "
              "this run will be ruled as a minimum! APOSMM may start a new run "
              "from some point in this run.")
        opt_flag = 0
    else:
        print("[APOSMM] NLopt returned with a negative return value, which indicates an error")
        opt_flag = 0

    if user_specs.get('periodic'):
        # Shift x_opt to be in the correct location in the unit cube
        # (not the domain user_specs['lb'] - user_specs['ub'])
        x_opt = x_opt % 1

    finish_queue(x_opt, opt_flag, comm_queue, parent_can_read, user_specs)


def run_local_scipy_opt(user_specs, comm_queue, x0, f0, child_can_read, parent_can_read):
    """
    Runs a SciPy local optimization run starting at ``x0``, governed by the
    parameters in ``user_specs``.
    """

    # Construct the bounds in the form of constraints
    cons = []
    for factor in range(len(x0)):
        lo = {'type': 'ineq',
              'fun': lambda x, lb=user_specs['lb'][factor], i=factor: x[i]-lb}
        up = {'type': 'ineq',
              'fun': lambda x, ub=user_specs['ub'][factor], i=factor: ub-x[i]}
        cons.append(lo)
        cons.append(up)

    method = user_specs['localopt_method'][6:]
    jac_flag = method in ['BFGS']
    # print('[Child]: Started my optimization', flush=True)
    res = sp_opt.minimize(lambda x: scipy_dfols_callback_fun(x, comm_queue,
                          child_can_read, parent_can_read, user_specs), x0,
                          # constraints=cons,
                          method=method, jac=jac_flag, **user_specs.get('scipy_kwargs', {}))

    if res['status'] in user_specs['opt_return_codes']:
        opt_flag = 1
    else:
        print("[APOSMM] The SciPy localopt run started from " + str(x0) + " stopped"
              " without finding a local min.\nThe 'status' of the run is " + str(res['status']) +
              " and the message is: \"" + res['message'] +
              "\".\nNo point from this run will be ruled as a minimum! APOSMM may "
              "start a new run from some point in this run.\n")
        opt_flag = 0

    if user_specs.get('periodic'):
        x_opt = res['x'] % 1
    else:
        x_opt = res['x']

    finish_queue(x_opt, opt_flag, comm_queue, parent_can_read, user_specs)


def run_external_localopt(user_specs, comm_queue, x0, f0, child_can_read, parent_can_read):
    """
    Runs an external local optimization run starting at ``x0``, governed by the
    parameters in ``user_specs``.
    """

    import subprocess
    import os
    from uuid import uuid4

    run_id = uuid4().hex

    x_file = 'x_' + run_id + '.txt'
    y_file = 'y_' + run_id + '.txt'
    x_done_file = 'x_done_' + run_id + '.txt'
    y_done_file = 'y_done_' + run_id + '.txt'
    opt_file = 'opt_' + run_id + '.txt'

    # cmd = ["matlab", "-nodisplay", "-nodesktop", "-nojvm", "-nosplash", "-r",
    cmd = ["octave", "--no-window-system", "--eval",
           "x0=[" + " ".join(["{:18.18f}".format(x) for x in x0]) + "];"
           "opt_file='" + opt_file + "';"
           "x_file='" + x_file + "';"
           "y_file='" + y_file + "';"
           "x_done_file='" + x_done_file + "';"
           "y_done_file='" + y_done_file + "';"
           "call_matlab_octave_script"]

    p = subprocess.Popen(cmd, shell=False, stdout=subprocess.DEVNULL)

    while p.poll() is None:  # Process still going
        if os.path.isfile(x_done_file):  # x file exists
            x = np.loadtxt(x_file)
            os.remove(x_done_file)

            x_recv, f_recv = put_set_wait_get(x, comm_queue, parent_can_read, child_can_read, user_specs)

            np.savetxt(y_file, [f_recv])
            open(y_done_file, 'w').close()

    x_opt = np.loadtxt(opt_file)
    opt_flag = np.loadtxt(opt_file + "_flag")

    for f in [x_file, y_file, opt_file]:
        os.remove(f)

    finish_queue(x_opt, opt_flag, comm_queue, parent_can_read, user_specs)


def run_local_dfols(user_specs, comm_queue, x0, f0, child_can_read, parent_can_read):
    """
    Runs a DFOLS local optimization run starting at ``x0``, governed by the
    parameters in ``user_specs``.
    """

    # Define bound constraints (lower <= x <= upper)
    lb = np.zeros(len(x0))
    ub = np.ones(len(x0))

    # Set random seed (for reproducibility)
    np.random.seed(0)

    # Care must be taken here because a too-large initial step causes DFO-LS to move the starting point!
    dist_to_bound = min(min(ub-x0), min(x0-lb))
    assert dist_to_bound > np.finfo(np.float64).eps, "The distance to the boundary is too small"
    assert 'bounds' not in user_specs.get('dfols_kwargs', {}), "APOSMM must set the bounds for DFO-LS"
    assert 'rhobeg' not in user_specs.get('dfols_kwargs', {}), "APOSMM must set rhobeg for DFO-LS"
    assert 'x0' not in user_specs.get('dfols_kwargs', {}), "APOSMM must set x0 for DFO-LS"

    # Call DFO-LS
    soln = dfols.solve(lambda x: scipy_dfols_callback_fun(x, comm_queue, child_can_read, parent_can_read, user_specs),
                       x0, bounds=(lb, ub), rhobeg=0.5*dist_to_bound, **user_specs.get('dfols_kwargs', {}))

    x_opt = soln.x

    if soln.flag == soln.EXIT_SUCCESS:
        opt_flag = 1
    else:
        print("[APOSMM] The DFO-LS run started from " + str(x0) + " stopped with an exit "
              "flag of " + str(soln.flag) + ". No point from this run will be "
              "ruled as a minimum! APOSMM may start a new run from some point "
              "in this run.")
        opt_flag = 0

    finish_queue(x_opt, opt_flag, comm_queue, parent_can_read, user_specs)


def run_local_tao(user_specs, comm_queue, x0, f0, child_can_read, parent_can_read):
    """
    Runs a PETSc/TAO local optimization run starting at ``x0``, governed by the
    parameters in ``user_specs``.
    """

    assert isinstance(x0, np.ndarray)

    tao_comm = PETSc.COMM_SELF
    n, = x0.shape
    if f0.shape == ():
        m = 1
    else:
        m, = f0.shape

    # Create starting point, bounds, and tao object
    x = PETSc.Vec().create(tao_comm)
    x.setSizes(n)
    x.setFromOptions()
    x.array = x0
    lb = x.duplicate()
    ub = x.duplicate()
    lb.array = 0*np.ones(n)
    ub.array = 1*np.ones(n)
    tao = PETSc.TAO().create(tao_comm)
    tao.setType(user_specs['localopt_method'])

    if user_specs['localopt_method'] == 'pounders':
        f = PETSc.Vec().create(tao_comm)
        f.setSizes(m)
        f.setFromOptions()

        if hasattr(tao, 'setResidual'):
            tao.setResidual(lambda tao, x, f: tao_callback_fun_pounders(tao, x, f, comm_queue, child_can_read,
                                                                        parent_can_read, user_specs), f)
        else:
            tao.setSeparableObjective(lambda tao, x, f: tao_callback_fun_pounders(tao, x, f, comm_queue, child_can_read,
                                                                                  parent_can_read, user_specs), f)
        delta_0 = user_specs['dist_to_bound_multiple']*np.min([np.min(ub.array-x.array), np.min(x.array-lb.array)])
        PETSc.Options().setValue('-tao_pounders_delta', str(delta_0))

    elif user_specs['localopt_method'] == 'nm':
        tao.setObjective(lambda tao, x: tao_callback_fun_nm(tao, x, comm_queue, child_can_read,
                                                            parent_can_read, user_specs))

    elif user_specs['localopt_method'] == 'blmvm':
        g = PETSc.Vec().create(tao_comm)
        g.setSizes(n)
        g.setFromOptions()
        tao.setObjectiveGradient(lambda tao, x, g: tao_callback_fun_grad(tao, x, g, comm_queue, child_can_read,
                                                                         parent_can_read, user_specs))

    # Set everything for tao before solving
    PETSc.Options().setValue('-tao_max_funcs', str(user_specs.get('run_max_eval', 1000*n)))
    tao.setFromOptions()
    tao.setVariableBounds((lb, ub))

    tao.setTolerances(grtol=user_specs.get('grtol', 1e-8), gatol=user_specs.get('gatol', 1e-8))
    tao.setInitial(x)

    # print('[Child]: Started my optimization', flush=True)
    tao.solve(x)

    x_opt = tao.getSolution().getArray()
    exit_code = tao.getConvergedReason()

    if exit_code > 0:
        opt_flag = 1
    else:
        # https://www.mcs.anl.gov/petsc/petsc-current/docs/manualpages/Tao/TaoGetConvergedReason.html
        print("[APOSMM] The run started from " + str(x0) + " exited with a nonpositive reason. No point from "
              "this run will be ruled as a minimum! APOSMM may start a new run from some point in this run.")
        opt_flag = 0

    if user_specs['localopt_method'] == 'pounders':
        f.destroy()
    elif user_specs['localopt_method'] == 'blmvm':
        g.destroy()

    lb.destroy()
    ub.destroy()
    x.destroy()
    tao.destroy()

    finish_queue(x_opt, opt_flag, comm_queue, parent_can_read, user_specs)


def opt_runner(run_local_opt, user_specs, comm_queue, x0, f0, child_can_read, parent_can_read):
    try:
        run_local_opt(user_specs, comm_queue, x0, f0, child_can_read, parent_can_read)
    except Exception as e:
        comm_queue.put(ErrorMsg(e))
        parent_can_read.set()


# Callback functions and routines
def nlopt_callback_fun(x, grad, comm_queue, child_can_read, parent_can_read, user_specs):

    if user_specs['localopt_method'] in ['LD_MMA']:
        x_recv, f_recv, grad_recv = put_set_wait_get(x, comm_queue, parent_can_read, child_can_read, user_specs)
        grad[:] = grad_recv
    else:
        assert user_specs['localopt_method'] in ['LN_SBPLX', 'LN_BOBYQA', 'LN_NEWUOA',
                                                 'LN_COBYLA', 'LN_NELDERMEAD', 'LD_MMA']
        x_recv, f_recv = put_set_wait_get(x, comm_queue, parent_can_read, child_can_read, user_specs)

    return f_recv


def scipy_dfols_callback_fun(x, comm_queue, child_can_read, parent_can_read, user_specs):

    if user_specs['localopt_method'] in ['scipy_BFGS']:

        x_recv, f_x_recv, grad_recv, = put_set_wait_get(x, comm_queue, parent_can_read, child_can_read, user_specs)

        return f_x_recv, grad_recv

    x_recv, f_x_recv, = put_set_wait_get(x, comm_queue, parent_can_read, child_can_read, user_specs)

    return f_x_recv


def tao_callback_fun_nm(tao, x, comm_queue, child_can_read, parent_can_read, user_specs):

    x_recv, f_recv, = put_set_wait_get(x.array_r, comm_queue, parent_can_read, child_can_read, user_specs)

    return f_recv


def tao_callback_fun_pounders(tao, x, f, comm_queue, child_can_read, parent_can_read, user_specs):

    x_recv, f_recv, = put_set_wait_get(x.array_r, comm_queue, parent_can_read, child_can_read, user_specs)
    f.array[:] = f_recv

    return f


def tao_callback_fun_grad(tao, x, g, comm_queue, child_can_read, parent_can_read, user_specs):

    x_recv, f_recv, grad_recv = put_set_wait_get(x.array_r, comm_queue, parent_can_read, child_can_read, user_specs)
    g.array[:] = grad_recv

    return f_recv


def finish_queue(x_opt, opt_flag, comm_queue, parent_can_read, user_specs):

    if user_specs.get('print') and opt_flag:
        print('[APOSMM] Local optimum on the [0,1]^n domain', x_opt, flush=True)
    comm_queue.put(ConvergedMsg(x_opt, opt_flag))
    parent_can_read.set()


def put_set_wait_get(x, comm_queue, parent_can_read, child_can_read, user_specs):
    """This routine is used by children process callback functions. It:
    - puts x into a comm_queue,
    - tells the parent it can read,
    - tells the child to wait
    - receives the values put in the comm_queue by the parent
    - checks that the first value received matches x
    - removes the wait on the child
    - returns values"""

    comm_queue.put(x)
    # print('[Child]: I just put x_on_cube =', x.array, flush=True)
    # print('[Child]: Parent should no longer wait.', flush=True)
    parent_can_read.set()
    # print('[Child]: I have started waiting', flush=True)
    child_can_read.wait()
    # print('[Child]: Wohooo.. I am free folks', flush=True)
    values = comm_queue.get()
    child_can_read.clear()

    if user_specs.get('periodic'):
        assert np.allclose(x % 1, values[0] % 1, rtol=1e-15, atol=1e-15), "The point I gave is not the point I got back"
    else:
        assert np.allclose(x, values[0], rtol=1e-15, atol=1e-15), "The point I gave is not the point I got back"

    return values


def simulate_recv_from_manager(local_H, gen_specs):
    # This function goes through any entries of local_H and if they have not
    # "returned", then it performs all function/gradient evaluations and makes
    # output as if the calculations were performed externally by libEnsemble.
    user = gen_specs['user']['standalone']

    if np.sum(local_H['returned']) >= user['eval_max']:
        return STOP_TAG, {}, {}

    H_rows = np.where(~local_H['returned'])[0]
    H_fields = [i[0] for i in gen_specs['out']]

    Work = {'libE_info': {'H_rows': H_rows}, 'H_fields': H_fields}

    calc_in = np.zeros(len(H_rows), dtype=gen_specs['out'] + [('f', float), ('grad', float, len(local_H['x'][0]))])

    for name in H_fields:
        calc_in[name] = local_H[name][H_rows]

    assert 'obj_func' in user or 'obj_and_grad_func' in user, "Must have some way to calculate objective values"

    if 'obj_func' in user:
        for i, row in enumerate(H_rows):
            calc_in['f'][i] = user['obj_func'](local_H['x'][row])

        if 'grad' in local_H.dtype.names:
            for i, row in enumerate(H_rows):
                calc_in['grad'][i] = user['grad_func'](local_H['x'][row])
    else:
        for i, row in enumerate(H_rows):
            out = user['obj_and_grad_func'](local_H['x'][row])
            calc_in['f'][i] = out[0]
            calc_in['grad'][i] = out[1]

    return EVAL_GEN_TAG, Work, calc_in



"""
Runs libEnsemble with uniform random sampling on a simple 1D problem
without specifying the default communicator. Tests that mpi is taken
as default, with a duplicate of MPI.COMM_WORLD. If libEnsemble
uses MPI.COMM_WORLD, this test will fail.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_1d_uniform_sampling_with_comm_dup.py

The number of concurrent evaluations of the objective function will be 4-1=3.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi
# TESTSUITE_NPROCS: 2 4

import sys
import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.one_d_func import one_d_example as sim_f
from libensemble.gen_funcs.sampling import uniform_random_sample as gen_f
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()

if libE_specs['comms'] != 'mpi':
    sys.exit("This test only runs with MPI -- aborting...")
    # After this check will remove libE_specs['comms']
else:
    from mpi4py import MPI

libE_specs = None  # Let MPI use defaults

# Check independence of default communicator from MPI.COMM_WORLD
world = MPI.COMM_WORLD
if not is_manager:
    world.isend(world.Get_rank(), dest=0, tag=0)

sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
}

gen_specs = {
    'gen_f': gen_f,
    'in': ['sim_id'],
    'out': [('x', float, (1,))],
    'user': {
        'lb': np.array([-3]),
        'ub': np.array([3]),
        'gen_batch_size': 500,
    },
}

persis_info = add_unique_random_streams({}, nworkers + 1)

exit_criteria = {'gen_max': 501}

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, libE_specs=libE_specs)

if is_manager:
    # assert libE_specs['comms'] == 'mpi', 'MPI default comms should be set'
    # Potential to cause a hang
    worker_ids = []
    exp_worker_ids = list(range(1, nworkers + 1))
    mpi_status = MPI.Status()
    for worker in range(1, nworkers + 1):
        worker_rank = world.recv(source=worker, status=mpi_status)
        worker_ids.append(worker_rank)
    assert worker_ids == exp_worker_ids, 'MPI World values are not as expected'
    assert len(H) >= 501
    print("\nlibEnsemble with random sampling has generated enough points")
    save_libE_output(H, persis_info, __file__, nworkers)



"""
This module contains a test noisy function
"""

import numpy as np
from numpy.linalg import norm
from numpy import cos, sin


def func_wrapper(H, persis_info, sim_specs, libE_info):
    """
    Wraps an objective function

    .. seealso::
        `test_persistent_fd_param_finder.py` <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_persistent_fd_param_finder.py>`_ # noqa
    """

    batch = len(H['x'])
    H0 = np.zeros(batch, dtype=sim_specs['out'])

    for i, x in enumerate(H['x']):
        H0['f_val'][i] = noisy_function(x)[H['f_ind'][i]]

    return H0, persis_info


def noisy_function(x):
    """
    """

    x1 = x[0]
    x2 = x[1]
    term1 = (4-2.1*x1**2+(x1**4)/3) * x1**2
    term2 = x1*x2
    term3 = (-4+4*x2**2) * x2**2

    phi1 = 0.9*sin(100*norm(x, 1))*cos(100*norm(x, np.inf)) + 0.1*cos(norm(x, 2))
    phi1 = phi1*(4*phi1**2 - 3)

    phi2 = 0.8*sin(100*norm(x, 1))*cos(100*norm(x, np.inf)) + 0.2*cos(norm(x, 2))
    phi2 = phi2*(4*phi2**2 - 3)

    phi3 = 0.7*sin(100*norm(x, 1))*cos(100*norm(x, np.inf)) + 0.3*cos(norm(x, 2))
    phi3 = phi3*(4*phi3**2 - 3)

    F = np.zeros(3)
    F[0] = (1 + 1e-1*phi1)*term1
    F[1] = (1 + 1e-2*phi2)*term2
    F[2] = (1 + 1e-3*phi3)*term3

    return F



from libensemble.alloc_funcs.give_sim_work_first import give_sim_work_first

alloc_specs = {
    'alloc_f': give_sim_work_first,
    'user': {
        'num_active_gens': 1}, }
# end_alloc_specs_rst_tag



import os
import shutil
import numpy as np
from libensemble.output_directory import EnsembleDirectory
from libensemble.utils.loc_stack import LocationStack


def test_range_single_element():
    """Single H_row labeling """

    work = {'H_fields': ['x', 'num_nodes', 'procs_per_node'],
            'libE_info': {'H_rows': np.array([5]), 'workerID': 1}}
    assert EnsembleDirectory.extract_H_ranges(work) == '5', \
        'Failed to correctly parse single H row'


def test_range_two_separate_elements():
    """Multiple H_rows, non-sequential"""

    work = {'H_fields': ['x', 'num_nodes', 'procs_per_node'],
            'libE_info': {'H_rows': np.array([2, 8]), 'workerID': 1}}
    assert EnsembleDirectory.extract_H_ranges(work) == '2_8', \
        'Failed to correctly parse nonsequential H rows'


def test_range_two_ranges():
    """Multiple sequences of H_rows"""

    work = {'H_fields': ['x', 'num_nodes', 'procs_per_node'],
            'libE_info': {'H_rows': np.array([0, 1, 2, 3, 7, 8]),
                          'workerID': 1}}
    assert EnsembleDirectory.extract_H_ranges(work) == '0-3_7-8', \
        'Failed to correctly parse multiple H ranges'


def test_range_mixes():
    """Mix of single rows and sequences of H_rows"""

    work = {'H_fields': ['x', 'num_nodes', 'procs_per_node'],
            'libE_info': {'H_rows': np.array([2, 3, 4, 6, 8, 9, 11, 14]),
                          'workerID': 1}}
    assert EnsembleDirectory.extract_H_ranges(work) == '2-4_6_8-9_11_14', \
        'Failed to correctly parse H row single elements and ranges.'


def test_copy_back():
    """ When workers conclude their work, workers have the option of copying
    back their work into a directory created by the manager."""

    inputdir = './calc'
    copybackdir = './calc_back'
    inputfile = './calc/file'

    for dir in [inputdir, copybackdir]:
        os.makedirs(dir, exist_ok=True)

    libE_specs = {'sim_dirs_make': True, 'ensemble_dir_path': inputdir,
                  'ensemble_copy_back': True}

    ls = LocationStack()
    ls.register_loc('test', inputfile)
    # fake_worker = FakeWorker(libE_specs, inputdir, '.', ls)
    ed = EnsembleDirectory(libE_specs, ls)
    EnsembleDirectory.copy_back(ed)
    assert 'file' in os.listdir(copybackdir), \
        'File not copied back to starting dir'

    for dir in [inputdir, copybackdir]:
        shutil.rmtree(dir)


def test_worker_dirs_but_no_sim_dirs():
    """Test Worker._make_calc_dir() directory structure without sim_dirs"""
    inputdir = './calc'
    inputfile = './calc/file'
    ensemble_dir = './test_ens'

    for dir in [inputdir, inputfile, ensemble_dir]:
        os.makedirs(dir, exist_ok=True)

    libE_specs = {'ensemble_dir_path': ensemble_dir, 'use_worker_dirs': True, 'sim_input_dir': inputdir}

    ls = LocationStack()
    ed = EnsembleDirectory(libE_specs, ls)
    for i in range(4):  # Should work at any range
        EnsembleDirectory._make_calc_dir(ed, 1, 1, 'sim', ls)

    assert 'worker1' in os.listdir(ensemble_dir)
    assert 'file' in os.listdir(os.path.join(ensemble_dir, 'worker1'))

    for dir in [inputdir, ensemble_dir]:
        shutil.rmtree(dir)


def test_loc_stack_FileExists_exceptions():
    inputdir = './calc'
    copyfile = './calc/copy'
    symlinkfile = './calc/symlink'
    ensemble_dir = './test_ens'

    for dir in [inputdir, copyfile, symlinkfile]:
        os.makedirs(dir, exist_ok=True)

    # Testing loc_stack continuing on FileExistsError when not using sim_dirs
    libE_specs = {'sim_dirs_make': False, 'ensemble_dir_path': ensemble_dir,
                  'use_worker_dirs': True, 'sim_dir_copy_files': [copyfile],
                  'sim_dir_symlink_files': [symlinkfile]}

    ls = LocationStack()
    ed = EnsembleDirectory(libE_specs=libE_specs, loc_stack=ls)
    for i in range(2):  # Should work at any range
        EnsembleDirectory._make_calc_dir(ed, 1, '1', 'sim', ls)

    assert len(os.listdir(ensemble_dir)) == 1, 'Should only be a single worker file in ensemble'
    assert 'worker1' in os.listdir(ensemble_dir), 'Only directory should be worker1'
    assert all([i in os.listdir(os.path.join(ensemble_dir, 'worker1')) for i in ['copy', 'symlink']]), \
        'Files to copy and symlink not found in worker directory.'

    # Testing loc_stack exception raising when sim_dir re-used - copy
    libE_specs = {'sim_dirs_make': True, 'ensemble_dir_path': ensemble_dir,
                  'use_worker_dirs': True, 'sim_dir_copy_files': [copyfile]}

    flag = 1
    ed = EnsembleDirectory(libE_specs=libE_specs, loc_stack=ls)
    EnsembleDirectory._make_calc_dir(ed, 1, '1', 'sim', ls)
    try:
        EnsembleDirectory._make_calc_dir(ed, 1, '1', 'sim', ls)
    except FileExistsError:
        flag = 0
    assert flag == 0

    # Testing loc_stack exception raising when sim_dir re-used - symlink
    libE_specs = {'sim_dirs_make': True, 'ensemble_dir_path': ensemble_dir,
                  'use_worker_dirs': True, 'sim_dir_symlink_files': [symlinkfile]}

    flag = 1
    ed = EnsembleDirectory(libE_specs=libE_specs, loc_stack=ls)
    EnsembleDirectory._make_calc_dir(ed, 1, '2', 'sim', ls)
    try:
        EnsembleDirectory._make_calc_dir(ed, 1, '2', 'sim', ls)
    except FileExistsError:
        flag = 0
    assert flag == 0

    for dir in [inputdir, ensemble_dir]:
        shutil.rmtree(dir)


if __name__ == '__main__':
    test_range_single_element()
    test_range_two_separate_elements()
    test_range_two_ranges()
    test_range_mixes()
    test_copy_back()
    test_worker_dirs_but_no_sim_dirs()
    test_loc_stack_FileExists_exceptions()



"""
This module contains an example 1d function
"""
__all__ = ['one_d_example']

import numpy as np


def one_d_example(x, persis_info, sim_specs, _):
    """
    Evaluates the six hump camel function for a single point ``x``.

    .. seealso::
        `test_fast_alloc.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_fast_alloc.py>`_ # noqa
    """

    H_o = np.zeros(1, dtype=sim_specs['out'])

    H_o['f'] = np.linalg.norm(x)

    return H_o, persis_info



import numpy as np


def six_hump_camel(H, persis_info, sim_specs, _):
    """Six-Hump Camel sim_f."""

    batch = len(H['x'])                            # Num evaluations each sim_f call.
    H_o = np.zeros(batch, dtype=sim_specs['out'])  # Define output array H

    for i, x in enumerate(H['x']):
        H_o['f'][i] = six_hump_camel_func(x)     # Function evaluations placed into H

    return H_o, persis_info


def six_hump_camel_func(x):
    """ Six-Hump Camel function definition """
    x1 = x[0]
    x2 = x[1]
    term1 = (4-2.1*x1**2+(x1**4)/3) * x1**2
    term2 = x1*x2
    term3 = (-4+4*x2**2) * x2**2

    return term1 + term2 + term3



#!/usr/bin/env python

"""
Unit test of termination with comms

Note that this must come before anything that loads PETSc, since PETSc
installs a SIGTERM handler.  Temporarily uninstalled from the test suite,
since pytest slurps up everything (including all the modules) in one go.
"""

import time
import signal
from libensemble.tools.tools import osx_set_mp_method
import libensemble.comms.comms as comms

osx_set_mp_method()


def test_qcomm_proc_terminate1():
    "Test that an already-done QCommProcess gracefully handles terminate()."

    def worker_main(comm):
        return

    with comms.QCommProcess(worker_main) as mgr_comm:
        time.sleep(0.5)
        mgr_comm.terminate(timeout=30)
        assert not mgr_comm.running


def test_qcomm_proc_terminate2():
    "Test that a QCommProcess run amok can be gracefully terminated."

    def worker_main(comm):
        while True:
            time.sleep(1)

    with comms.QCommProcess(worker_main) as mgr_comm:
        mgr_comm.terminate(timeout=30)
        assert not mgr_comm.running


def ignore_handler(signum, frame):
    print("Ignoring SIGTERM")


def test_qcomm_proc_terminate3():
    "Test that a QCommProcess ignoring SIGTERM manages."

    def worker_main(comm):
        signal.signal(signal.SIGTERM, ignore_handler)
        while not comm.mail_flag():
            pass

    with comms.QCommProcess(worker_main) as mgr_comm:
        time.sleep(0.5)

        flag = True
        try:
            mgr_comm.recv(timeout=0.5)
            flag = False
        except comms.Timeout:
            pass
        assert flag, "Should time out on recv"

        flag = True
        try:
            mgr_comm.result(timeout=0.5)
            flag = False
        except comms.Timeout:
            pass
        assert flag, "Should time out on result"

        flag = True
        try:
            mgr_comm.terminate(timeout=1)
            flag = False
        except comms.Timeout:
            pass
        assert flag, "Should time out on terminate"

        assert mgr_comm.running, "Should still be running"
        mgr_comm.send("Done")


def test_qcomm_proc_terminate4():
    "Test that a QCommProcess can handle event timeouts correctly."

    def worker_main(comm):
        while not comm.mail_flag():
            comm.send("Hello")
            time.sleep(0.01)

    with comms.QCommProcess(worker_main) as mgr_comm:
        time.sleep(0.5)

        flag = True
        try:
            mgr_comm.result(timeout=0.5)
            flag = False
        except comms.Timeout:
            pass
        assert flag, "Should time out on result"

        assert mgr_comm.running, "Should still be running"
        mgr_comm.send("Done")



# Test of executor running multiple tasks for libensemble. Could support
# hybrid mode - including, eg. running multi tasks per node (launched locally),
# or simply sharing burden on central system/consecutive pipes to balsam
# database - could enable use of threads if supply run-directories rather than
# assuming in-place runs etc....

# Test does not require running full libensemble
import os
from libensemble.executors.executor import Executor


def build_simfunc():
    import subprocess

    # Build simfunc
    # buildstring='mpif90 -o my_simtask.x my_simtask.f90' # On cray need to use ftn
    buildstring = 'mpicc -o my_simtask.x simdir/my_simtask.c'
    # subprocess.run(buildstring.split(),check=True) # Python3.5+
    subprocess.check_call(buildstring.split())


# --------------- Calling script ---------------------------------------------------------------
# sim_app = 'simdir/my_simtask.x'
# gen_app = 'gendir/my_genjob.x'

# temp
sim_app = './my_simtask.x'

if not os.path.isfile(sim_app):
    build_simfunc()

USE_BALSAM = False  # Take as arg
# USE_BALSAM = True # Take as arg

# Create and add exes to registry
if USE_BALSAM:
    from libensemble.baslam_executor import BalsamMPIExecutor
    exctr = BalsamMPIExecutor()
else:
    from libensemble.executors.mpi_executor import MPIExecutor
    exctr = MPIExecutor()

exctr.register_app(full_path=sim_app, calc_type='sim')

# Alternative to IF could be using eg. fstring to specify: e.g:
# EXECUTOR = 'Balsam'
# registry = f"{EXECUTOR}Register()"

# --------------- Worker: sim func -------------------------------------------------------------
# Should work with Balsam or not

# Can also use an internal iterable list of tasks in EXECUTOR - along with all_done func etc...


def polling_loop(exctr, task_list, timeout_sec=40.0, delay=1.0):
    import time
    start = time.time()

    while time.time() - start < timeout_sec:

        # Test all done - (return list of not-finished tasks and test if empty)
        active_list = [task for task in task_list if not task.finished]
        if not active_list:
            break

        for task in task_list:
            if not task.finished:
                time.sleep(delay)
                print('Polling task {0} at time {1}'.
                      format(task.id, time.time() - start))
                task.poll()

                if task.finished:
                    continue
                elif task.state == 'WAITING':
                    print('Task {0} waiting to execute'.format(task.id))
                elif task.state == 'RUNNING':
                    print('Task {0} still running ....'.format(task.id))

                # Check output file for error
                if task.stdout_exists():
                    if 'Error' in task.read_stdout():
                        print("Found (deliberate) Error in output file - "
                              "cancelling task {}".format(task.id))
                        exctr.kill(task)
                        time.sleep(delay)  # Give time for kill
                        continue

                # But if I want to do something different -
                #  I want to make a file - no function for THAT!
                # But you can get all the task attributes!
                # Uncomment to test
                # path = os.path.join(task.workdir,'newfile'+str(time.time()))
                # open(path, 'a')

    print('Loop time', time.time() - start)

    for task in task_list:
        if task.finished:
            if task.state == 'FINISHED':
                print('Task {0} finished successfully. Status: {1}'.
                      format(task.id, task.state))
            elif task.state == 'FAILED':
                print('Task {0} failed. Status: {1}'.
                      format(task.id, task.state))
            elif task.state == 'USER_KILLED':
                print('Task {0} has been killed. Status: {1}'.
                      format(task.id, task.state))
            else:
                print('Task {0} status: {1}'.format(task.id, task.state))
        else:
            print('Task {0} timed out. Status: {1}'.format(task.id, task.state))
            exctr.kill(task)
            if task.finished:
                print('Task {0} Now killed. Status: {1}'.
                      format(task.id, task.state))
                # double check
                task.poll()
                print('Task {0} state is {1}'.format(task.id, task.state))


# Tests

# From worker call Executor by different name to ensure getting registered
# app from Executor
exctr = Executor.executor


print('\nTest 1 - 3 tasks should complete successfully with status FINISHED :\n')

task_list = []
cores = 4

for j in range(3):
    # Could allow submission to generate outfile names based on task.id
    # outfilename = 'out_' + str(j) + '.txt'
    sleeptime = 6 + j*3  # Change args
    args_for_sim = 'sleep' + ' ' + str(sleeptime)
    rundir = 'run_' + str(sleeptime)
    task = exctr.submit(calc_type='sim', num_procs=cores, app_args=args_for_sim)
    task_list.append(task)


polling_loop(exctr, task_list)



"""
Runs libEnsemble on a 3-objective problem using the VTMOP package.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_VTMOP.py
   python3 test_VTMOP.py --nworkers 3 --comms local
   python3 test_VTMOP.py --nworkers 3 --comms tcp

This requires that VTMOP be installed. Contact thchang@vt.edu for a copy.
To build, use  : $ make genfuncs
To add to path : $ export PATH=$PATH:`pwd` (from src/build directory)

The wrapper for VTMOP is in gen_funcs/vtmop.py
Running this test will generate 3 unformatted binary files mop.io, mop.dat,
and mop.chkpt in the working directory, for sharing data between VTMOP
and libE.

The number of concurrent evaluations of the objective function will be 4-1=3.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: local
# TESTSUITE_OS_SKIP: OSX
# TESTSUITE_NPROCS: 5
# TESTSUITE_EXTRA: true

import numpy as np
import os
import time
from libensemble.utils.timer import Timer

# Import libEnsemble items for this test
from libensemble.libE import libE

# from libensemble.sim_funcs.mop_funcs import convex_mop as func
from libensemble.sim_funcs.mop_funcs import dtlz2 as func
from libensemble.gen_funcs.vtmop import vtmop_gen as gen_f
from libensemble.alloc_funcs.only_one_gen_alloc import ensure_one_active_gen as alloc_f
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams

timer = Timer()

# Set the problem dimensions here
num_dims = 5
num_objs = 3
lower = 0.0
upper = 1.0


# Definition of the sum function
def sim_f(H, *unused):
    # Initialize the output array
    Out = np.zeros(1, dtype=sim_specs['out'])
    # Evaluate the objective functions
    f = np.ones(np.size(Out['f']))
    func(H['x'][0], f)
    # Return the output array
    Out['f'] = f
    return Out, {}


# Set up the problem
nworkers, is_manager, libE_specs, _ = parse_args()
lower_bounds = lower * np.ones(num_dims)
upper_bounds = upper * np.ones(num_dims)

# Set up the simulator
sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float, num_objs)],
}

# Set up the generator
gen_specs = {
    'gen_f': gen_f,  # Set the gen to VTMOP (aliased to gen_f above).
    'in': ['x', 'f'],
    'out': [('x', float, num_dims)],
    'user': {
        # Set the problem dimensions (must match lb and ub).
        'd': num_dims,
        'p': num_objs,
        # Set the bound constraints.
        'lb': lower_bounds,
        'ub': upper_bounds,
        # Is this the beginning of a new run?
        'new_run': True,
        # isnb specifies the size of the initial search
        # and should generally be a large number. However, if a
        # precomputed database is available, then the initial search
        # could be skipped. If 0 is given, then the initial search is
        # skipped. Setting first_batch_size to 0 without supplying an
        # initial database will cause an error since the surrogates
        # cannot be fit without sufficient data.
        'isnb': 1000,
        # snb is the number of points used to search
        # each local trust region (using Latin hypercube design).
        # This should be a multiple of the number of concurrent
        # function evaluations and on the order of 4*d (where d is
        # the number of design variables)
        'snb': int(np.ceil(4 * num_dims / nworkers) * nworkers),
        # onb is the preferred number of candidate designs.
        # When the actual number of candidates is not a multiple of
        # opt_batch_size, additional candidates are randomly generated
        # to pad out the batch (if possible). This should be the exact
        # number of concurrent simulations used.
        'onb': nworkers,
        # Other optional arguments below:
        # Set the trust region radius as a fraction of ub[:]-lb[:].
        # This setting is problem dependent. A good starting place
        # would be between 0.1 and 0.2.
        'trust_radf': 0.1,
    },
}

# Set up the allocator
alloc_specs = {'alloc_f': alloc_f, 'out': []}

s1 = []
H = []

for run in range(3):
    if run == 0:
        # Run for 1100 evaluations or 300 seconds
        H0 = None
        exit_criteria = {'sim_max': 1100, 'elapsed_wallclock_time': 300}

    elif run == 1:
        # In the second run, we initialize VTMOP with an initial sample of previously evaluated points
        np.random.seed(0)
        size = 1000

        # Generate the sample
        X = np.random.uniform(gen_specs['user']['lb'], gen_specs['user']['ub'], (size, num_dims))
        f = np.zeros((size, num_objs))

        # Initialize H0
        H0_dtype = [
            ('x', float, num_dims),
            ('f', float, num_objs),
            ('sim_id', int),
            ('returned', bool),
            ('given', bool),
        ]
        H0 = np.zeros(size, dtype=H0_dtype)
        H0['x'] = X
        H0['sim_id'] = range(size)
        H0[['given', 'returned']] = True

        # Perform objective function evaluations
        for i in range(size):
            Out, _ = sim_f(H0[[i]])
            H0['f'][i] = Out['f']

        # Run for 200 more evaluations or 300 seconds
        exit_criteria = {'sim_max': 200, 'elapsed_wallclock_time': 300}

        gen_specs['user']['isnb'] = 0
        gen_specs['user']['new_run'] = True  # Need to set this as it can be overwritten within the libE call.

    elif run == 2:
        # In the third run, we restart VTMOP by loading in the history array saved in run==1
        gen_specs['user']['new_run'] = False

        # Inelegant way to have the manager copy over the VTMOP checkpoint
        # file, and have every worker get the H value from the run==1 case to
        # use in the restart.
        try:
            os.remove('manager_done_file')
        except OSError:
            pass

        if is_manager:
            os.rename('vtmop.chkpt_finishing_' + s1, 'vtmop.chkpt')
            np.save('H_for_vtmop_restart.npy', H)
            open('manager_done_file', 'w').close()
        else:
            while not os.path.isfile('manager_done_file'):
                time.sleep(0.1)
            H = np.load('H_for_vtmop_restart.npy')

        # Initialize H0 with values from H (from the run==1 case)
        size = sum(H['returned'])
        H0_dtype = [
            ('x', float, num_dims),
            ('f', float, num_objs),
            ('sim_id', int),
            ('returned', bool),
            ('given', bool),
        ]
        H0 = np.zeros(size, dtype=H0_dtype)
        H0['x'] = H['x'][:size]
        H0['sim_id'] = range(size)
        H0[['given', 'returned']] = True
        H0['f'] = H['f'][:size]

        # Run for 200 more evaluations or 300 seconds
        exit_criteria = {'sim_max': 200, 'elapsed_wallclock_time': 300}

    # Persistent info between iterations
    persis_info = add_unique_random_streams({}, nworkers + 1)
    persis_info['next_to_give'] = 0 if H0 is None else len(H0)
    persis_info['total_gen_calls'] = 0

    # Perform the run
    H, persis_info, flag = libE(
        sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs=alloc_specs, libE_specs=libE_specs, H0=H0
    )

    # The manager takes care of checkpointing/output
    if is_manager:
        # Renaming vtmop checkpointing file, if needed for later use.
        timer.start()
        s1 = timer.date_start.replace(' ', '_')
        os.rename('vtmop.chkpt', 'vtmop.chkpt_finishing_' + s1)

        assert flag == 0
        save_libE_output(H, persis_info, __file__, nworkers)



#!/usr/bin/env python

# infile='testfile.txt'
infile = 'libE_stats.txt'
time_key = 'Time:'
status_key = 'Status:'
sim_only = True  # Ignore generator times
num_bins = 40

# States - could add multiple lines - eg Failed.
ran_ok = ['Completed']  # list of ok states

# run_killed = ['killed', 'Exception'] # Currently searches for this word in string
run_killed = ['killed']  # Currently searches for this word in string
run_exception = ['Exception', 'Failed']

import sys
import numpy as np

import matplotlib
matplotlib.use('Agg')

import matplotlib.pyplot as plt
# import csv


def search_for_keyword(in_list, kw_list):
    for i, val in enumerate(in_list):
        if val.endswith(':'):
            break  # New key word found
        else:
            if val in kw_list:
                return True
    return False


def append_to_list(mylst, glob_list, found_time):
    # Assumes Time comes first - else have to modify
    if found_time:
        mylst.append(glob_list[-1])
    else:
        print('Error Status found before time - exiting')
        sys.exit()


active_line_count = 0
in_times = []
in_times_ran = []
in_times_kill = []
in_times_exception = []
exceptions = False


# Read straight from libEnsemble summary file.
with open(infile) as f:
    for line in f:
        lst = line.split()
        found_time = False
        found_status = False
        for i, val in enumerate(lst):
            if val == time_key:
                if sim_only and lst[i-1] != 'sim':
                    break
                in_times.append(lst[i+1])
                found_time = True
            if val == status_key:
                if lst[i+1] in ran_ok:
                    append_to_list(in_times_ran, in_times, found_time)  # Assumes Time comes first
                elif search_for_keyword(lst[i+1:len(lst)], run_killed):
                    append_to_list(in_times_kill, in_times, found_time)  # Assumes Time comes first
                elif search_for_keyword(lst[i+1:len(lst)], run_exception):
                    exceptions = True
                    append_to_list(in_times_exception, in_times, found_time)  # Assumes Time comes first
                else:
                    print('Error: Unknown status - rest of line: {}'.format(lst[i+1:len(lst)]))
                    sys.exit()
                found_status = True
            if found_time and found_status:
                active_line_count += 1
                break

print('Processed {} calcs'.format(active_line_count))

times = np.asarray(in_times, dtype=float)
times_ran = np.asarray(in_times_ran, dtype=float)
times_kill = np.asarray(in_times_kill, dtype=float)

binwidth = (times.max() - times.min()) / num_bins
bins = np.arange(min(times), max(times) + binwidth, binwidth)

# plt.hist(times, bins, histtype='bar', rwidth=0.8)
p1 = plt.hist(times_ran, bins, label='Completed')
p2 = plt.hist(times_kill, bins, label='Killed')

if exceptions:
    times_exc = np.asarray(in_times_exception, dtype=float)
    p3 = plt.hist(times_exc, bins, label='Except/Failed', color='C3')  # red

# plt.title('Theta Opal/libEnsemble Times: 127 Workers - sim_max 508')
if sim_only:
    calc = 'sim'
else:
    calc = 'calc'

titl = ('Histogram of ' + calc + ' times' + ' (' + str(active_line_count) + ' user calcs)' + str(num_bins) + ' bins')

plt.title(titl)
plt.xlabel('Calculation run time (sec)', fontsize=14)
plt.ylabel('Count', fontsize=14)
plt.grid(True)
plt.legend(loc='best', fontsize=14)

plt.xticks(fontsize=12)
plt.yticks(fontsize=12)
# plt.show()
plt.savefig('hist_completed_v_killed.png', bbox='tight', transparent=True)



import matplotlib
matplotlib.use('Agg')

from balsam.core import models
from matplotlib import pyplot as plt
import matplotlib.dates as mdates

fig, ax = plt.subplots()

plt.title('Balsam: Tasks completed v Time')
plt.xlabel('Time of Day (H:M)')
plt.ylabel('Num. Tasks Completed (Accum)')

times, throughputs = models.throughput_report()
plt.step(times, throughputs, where='post')

myFmt = mdates.DateFormatter('%H:%M')
ax.xaxis.set_major_formatter(myFmt)
fig.autofmt_xdate()

plt.savefig('balsam_tasks_v_time.png')



import re

"""
This file is part of the suite of scripts to use LibEnsemble on top of WarpX
simulations. It provides functions to modify some parameters in a WarpX
input file.
"""


def _set_value(str_text, str_line, val):
    """
    This function search for str_line in str_text and overwrites the whole
    matching line with <str_line> str(val)
    """
    str_text = re.sub('\n' + str_line + '.*',
                      '\n' + str_line + str(val),
                      str_text)
    return str_text


def write_sim_input(input_file, x_values):
    """
    This function modifies input_file to replace the values of some parameters
    by values provided in x_values.

    Parameters
    ----------
    input_file : WarpX input file.
    x_values : Specific value of input x for this run.
    """

    # Parameters exposed to optimization
    ramp_down_1 = x_values[0][0]
    ramp_down_2 = x_values[0][1]
    zlens_1 = x_values[0][2]
    adjust_factor = x_values[0][3]

    # Fixed parameters
    ramp_up_1 = 0.02
    plateau_1 = 0.297
    ramp_up_2 = ramp_up_1
    plateau_2 = plateau_1
    gap_12 = .0285

    end_stage_1 = ramp_up_1 + plateau_1 + ramp_down_1
    beg_stage_2 = end_stage_1 + gap_12
    end_stage_2 = beg_stage_2 + ramp_up_2 + plateau_2 + ramp_down_2

    # End simulation when beam has just escaped the last stage
    gamma_b = 30.
    zmax_stop_run = end_stage_2 - 55.e-6 * gamma_b**2 * 2.

    with open(input_file) as file_handler:
        output_text = file_handler.read()

    # Set end of stage 1
    output_text = _set_value(
        output_text, 'electrons.zmax = ', str(end_stage_1))
    output_text = _set_value(
        output_text, 'ions.zmax = ', str(end_stage_1))
    # Set length of final downramp of stage 1
    output_text = _set_value(
        output_text, 'electrons.predefined_profile_params = ',
        '0.0 .02 .297 ' + str(ramp_down_1) + ' 40.e-6 1.7e23')
    output_text = _set_value(
        output_text, 'ions.predefined_profile_params = ',
        '0.0 .02 .297 ' + str(ramp_down_1) + ' 40.e-6 1.7e23')
    # Set position of lens
    output_text = _set_value(
        output_text, 'my_constants.zlen = ', str(end_stage_1 + zlens_1))
    # Set beginning of stage 2
    output_text = _set_value(
        output_text, 'electrons2.zmin = ', str(beg_stage_2))
    output_text = _set_value(
        output_text, 'ions2.zmin = ', str(beg_stage_2))
    # Set end of stage 2
    output_text = _set_value(
        output_text, 'electrons2.zmax = ', str(end_stage_2))
    output_text = _set_value(
        output_text, 'ions2.zmax = ', str(end_stage_2))
    # Set length of final downramp of stage 2
    output_text = _set_value(
        output_text, 'electrons2.predefined_profile_params = ',
        str(beg_stage_2) + ' .02 .297 ' + str(ramp_down_2) + ' 40.e-6 1.7e23')
    output_text = _set_value(
        output_text, 'ions2.predefined_profile_params = ',
        str(beg_stage_2) + ' .02 .297 ' + str(ramp_down_2) + ' 40.e-6 1.7e23')
    # Set adjustment factor on lens strength
    output_text = _set_value(
        output_text, 'my_constants.adjust_factor = ', str(adjust_factor))
    # Set when to stop the run
    output_text = _set_value(
        output_text, 'warpx.zmax_plasma_to_compute_max_step = ',
        str(zmax_stop_run))

# Write new input file
    fout = open(input_file, 'w')
    fout.write(output_text)
    fout.close()



import numpy as np
import time
import logging

from libensemble.tools.fields_keys import libE_fields, protected_libE_fields

logger = logging.getLogger(__name__)

# For debug messages - uncomment
# logger.setLevel(logging.DEBUG)


class History:

    """The history class provides methods for managing the history array.

    **Object Attributes:**

    These are set on initialization.

    :ivar numpy_structured_array H:
        History array storing rows for each point. Field names are in
        libensemble/tools/fields_keys.py

    :ivar int offset:
        Starting index for this ensemble (after H0 read in)

    :ivar int index:
        Index where libEnsemble should start filling in H

    :ivar int given_count:
        Number of points given to sim functions (according to H)

    :ivar int returned_count:
        Number of points evaluated  (according to H)

    Note that index, given_count and returned_count reflect the total number of points
    in H and therefore include those prepended to H in addition to the current run.

    """

    # Not currently using libE_specs, persis_info - need to add parameters
    # def __init__(self, libE_specs, alloc_specs, sim_specs, gen_specs, exit_criteria, H0, persis_info):
    def __init__(self, alloc_specs, sim_specs, gen_specs, exit_criteria, H0):
        """
        Forms the numpy structured array that records everything from the
        libEnsemble run

        """
        L = exit_criteria.get('sim_max', 100)

        # Combine all 'out' fields (if they exist) in sim_specs, gen_specs, or alloc_specs
        specs = [sim_specs, alloc_specs, gen_specs]
        dtype_list = list(set(libE_fields + sum([k.get('out', []) for k in specs if k], [])))
        H = np.zeros(L + len(H0), dtype=dtype_list)  # This may be more history than is needed if H0 has un-given points

        if len(H0):
            # Prepend H with H0
            fields = H0.dtype.names

            for field in fields:
                H[field][:len(H0)] = H0[field]
                # for ind, val in np.ndenumerate(H0[field]):  # Works if H0[field] has arbitrary dimension but is slow
                #     H[field][ind] = val

            if 'given' not in fields:
                logger.manager_warning("Marking entries in H0 as having been 'given' and 'returned'")
                H['given'][:len(H0)] = 1
                H['returned'][:len(H0)] = 1
            elif 'returned' not in fields:
                logger.manager_warning("Marking entries in H0 as having been 'returned' if 'given'")
                H['returned'][:len(H0)] = H0['given']

            if 'sim_id' not in fields:
                logger.manager_warning("Assigning sim_ids to entries in H0")
                H['sim_id'][:len(H0)] = np.arange(0, len(H0))

        H['sim_id'][-L:] = -1
        H['given_time'][-L:] = np.inf
        H['last_given_time'][-L:] = np.inf
        H['last_given_back_time'][-L:] = np.inf

        self.H = H
        self.using_H0 = len(H0) > 0
        self.index = len(H0)
        self.grow_count = 0

        self.given_count = np.sum(H['given'])
        self.returned_count = np.sum(H['returned'])
        self.given_back_count = np.sum(H['given_back'])
        self.given_back_warned = False

        self.given_offset = self.given_count
        self.returned_offset = self.returned_count
        self.given_back_offset = self.given_back_count

    def update_history_f(self, D, safe_mode):
        """
        Updates the history after points have been evaluated
        """

        new_inds = D['libE_info']['H_rows']  # The list of rows (as a numpy array)
        returned_H = D['calc_out']

        for j, ind in enumerate(new_inds):
            for field in returned_H.dtype.names:
                if safe_mode:
                    assert field not in protected_libE_fields, "The field '" + field + "' is protected"
                if np.isscalar(returned_H[field][j]):
                    self.H[field][ind] = returned_H[field][j]
                else:
                    # len or np.size
                    H0_size = len(returned_H[field][j])
                    assert H0_size <= len(self.H[field][ind]),\
                        "History update Error: Too many values received for " + field
                    assert H0_size, "History update Error: No values in this field " + field
                    if H0_size == len(self.H[field][ind]):
                        self.H[field][ind] = returned_H[field][j]  # ref
                    else:
                        self.H[field][ind][:H0_size] = returned_H[field][j]  # Slice View

            self.H['returned'][ind] = True
            self.H['returned_time'][ind] = time.time()
            self.returned_count += 1

    def update_history_x_out(self, q_inds, sim_worker):
        """
        Updates the history (in place) when new points have been given out to be evaluated

        Parameters
        ----------
        q_inds: numpy array
            Row IDs for history array H

        sim_worker: integer
            Worker ID
        """
        q_inds = np.atleast_1d(q_inds)
        first_given_inds = ~self.H['given'][q_inds]
        t = time.time()

        self.H['given'][q_inds] = True
        self.H['given_time'][q_inds[first_given_inds]] = t
        self.H['last_given_time'][q_inds] = t
        self.H['sim_worker'][q_inds] = sim_worker

        self.given_count += len(q_inds)

    def update_history_to_gen(self, q_inds):
        """Updates the history (in place) when points are given back to the gen"""
        q_inds = np.atleast_1d(q_inds)
        t = time.time()

        if q_inds.size > 0:
            if np.all(self.H['returned'][q_inds]):
                self.H['given_back'][q_inds] = True

            elif np.any(self.H['returned'][q_inds]):  # sporadic returned points need updating
                for ind in q_inds[self.H['returned'][q_inds]]:
                    self.H['given_back'][ind] = True

            if self.using_H0 and not self.given_back_warned:
                logger.manager_warning(
                    "Giving entries in H0 back to gen. Marking entries in H0 as 'given_back' if 'returned'.")
                self.given_back_warned = True

            self.H['last_given_back_time'][q_inds] = t
            self.given_back_count += len(q_inds)

    def update_history_x_in(self, gen_worker, D, safe_mode):
        """
        Updates the history (in place) when new points have been returned from a gen

        Parameters
        ----------
        gen_worker: integer
            The worker who generated these points
        D: numpy array
            Output from gen_func
        """

        if len(D) == 0:
            return

        rows_remaining = len(self.H)-self.index

        if 'sim_id' not in D.dtype.names:
            # gen method must not be adjusting sim_id, just append to self.H
            num_new = len(D)

            if num_new > rows_remaining:
                self.grow_count = max(num_new-rows_remaining, 2*self.grow_count)
                self.grow_H(self.grow_count)

            update_inds = np.arange(self.index, self.index+num_new)
            self.H['sim_id'][self.index:self.index+num_new] = range(self.index, self.index+num_new)
        else:
            # gen method is building sim_id or adjusting values in existing sim_id rows.

            # Ensure there aren't any gaps in the generated sim_id values:
            assert np.all(np.in1d(np.arange(self.index, np.max(D['sim_id'])+1), D['sim_id'])),\
                "The generator function has produced sim_id that are not in order."

            num_new = len(np.setdiff1d(D['sim_id'], self.H['sim_id']))

            if num_new > rows_remaining:
                self.grow_count = max(num_new-rows_remaining, 2*self.grow_count)
                self.grow_H(self.grow_count)

            update_inds = D['sim_id']

        for field in D.dtype.names:
            if safe_mode:
                assert field not in protected_libE_fields, "The field '" + field + "' is protected"
            self.H[field][update_inds] = D[field]

        first_gen_inds = update_inds[self.H['gen_time'][update_inds] == 0]
        t = time.time()
        self.H['gen_time'][first_gen_inds] = t
        self.H['last_gen_time'][update_inds] = t
        self.H['gen_worker'][first_gen_inds] = gen_worker
        self.index += num_new

    def grow_H(self, k):
        """
        Adds k rows to H in response to gen_f producing more points than
        available rows in H.

        Parameters
        ----------
        k: integer
            Number of rows to add to H
        """
        H_1 = np.zeros(k, dtype=self.H.dtype)
        H_1['sim_id'] = -1
        H_1['given_time'] = np.inf
        H_1['last_given_time'] = np.inf
        H_1['last_given_back_time'] = np.inf
        self.H = np.append(self.H, H_1)

    # Could be arguments here to return different truncations eg. all done, given etc...
    def trim_H(self):
        """Returns truncated array"""
        return self.H[:self.index]



from libensemble.message_numbers import STOP_TAG, PERSIS_STOP, UNSET_TAG, EVAL_GEN_TAG, EVAL_SIM_TAG, calc_type_strings
import logging
logger = logging.getLogger(__name__)


class PersistentSupport:
    """A helper class to assist with writing persistent user functions."""

    def __init__(self, libE_info, calc_type):
        """
        Instantiate a new PersistentSupport instance

        :param libE_info: A dictionary containing information about this work request
        :param calc_type: Named integer giving calculation type - EVAL_GEN_TAG or EVAL_SIM_TAG

        """
        self.libE_info = libE_info
        self.comm = self.libE_info['comm']
        self.calc_type = calc_type
        assert self.calc_type in [EVAL_GEN_TAG, EVAL_SIM_TAG], \
            "User function value {} specifies neither a simulator nor generator.".format(self.calc_type)
        self.calc_str = calc_type_strings[self.calc_type]

    def send(self, output, calc_status=UNSET_TAG):
        """
        Send message from worker to manager.

        :param output: Output array to be sent to manager
        :param calc_status: Optional, Provides a task status

        :returns: None

        """
        if 'comm' in self.libE_info:
            # Need to make copy before remove comm as original could be reused
            libE_info = dict(self.libE_info)
            libE_info.pop('comm')
        else:
            libE_info = self.libE_info

        D = {'calc_out': output,
             'libE_info': libE_info,
             'calc_status': calc_status,
             'calc_type': self.calc_type
             }
        logger.debug('Persistent {} function sending data message to manager'.format(self.calc_str))
        self.comm.send(self.calc_type, D)

    def recv(self):
        """
        Receive message to worker from manager.

        :returns: message tag, Work dictionary, calc_in array

        """
        tag, Work = self.comm.recv()  # Receive meta-data or signal
        if tag in [STOP_TAG, PERSIS_STOP]:
            logger.debug('Persistent {} received signal {} from manager'.format(self.calc_str, tag))
            self.comm.push_to_buffer(tag, Work)
            return tag, Work, None
        else:
            logger.debug('Persistent {} received work request from manager'.format(self.calc_str))

        # Update libE_info
        # self.libE_info = Work['libE_info']

        # Only replace rows - keep same dictionary
        self.libE_info['H_rows'] = Work['libE_info']['H_rows']

        data_tag, calc_in = self.comm.recv()  # Receive work rows

        # Check for unexpected STOP (e.g. error between sending Work info and rows)
        if data_tag in [STOP_TAG, PERSIS_STOP]:
            logger.debug('Persistent {} received signal {} '.format(self.calc_str, tag) +
                         'from manager while expecting work rows')
            self.comm.push_to_buffer(data_tag, calc_in)
            return data_tag, calc_in, None  # calc_in is signal identifier

        logger.debug('Persistent {} received work rows from manager'.format(self.calc_str))
        return tag, Work, calc_in

    def send_recv(self, output, calc_status=UNSET_TAG):
        """
        Send message from worker to manager and receive response.

        :param output: Output array to be sent to manager
        :param calc_status: Optional, Provides a task status

        :returns: message tag, Work dictionary, calc_in array

        """
        self.send(output, calc_status)
        return self.recv()



"""
This module contains a periodic test function
"""

import numpy as np
from numpy import cos, sin


def func_wrapper(H, persis_info, sim_specs, libE_info):
    """
    Wraps an objective function
    """

    batch = len(H['x'])
    H0 = np.zeros(batch, dtype=sim_specs['out'])

    for i, x in enumerate(H['x']):
        H0['f'][i] = periodic_func(x)

    return H0, persis_info


def periodic_func(x):
    """
    This function is periodic
    """
    return sin(x[0])*cos(x[1])



#! /usr/bin/env python

# Configuring Balsam for libEnsemble consists of installing the latest version
#   of Balsam from source and moving Balsam-relevant tests to the regression_tests
#   directory. This way, run-tests won't run a non-relevant test if Balsam or the
#   necessary Python version aren't installed.

import sys
import os
import subprocess


def install_balsam():
    here = os.getcwd()
    os.chdir('../balsam/balsam-0.5.0')
    subprocess.check_call('pip install -e .'.split())
    os.chdir(here)


def move_test_balsam():
    current_dir_with_btest = './libensemble/tests/regression_tests/scripts_used_by_reg_tests/test_balsam_hworld.py'
    reg_dir_with_btest = './libensemble/tests/regression_tests/test_balsam_hworld.py'
    if not os.path.isfile(reg_dir_with_btest):
        os.rename(current_dir_with_btest, reg_dir_with_btest)


def configure_coverage():
    # Enables coverage of balsam_controller.py if running test
    coveragerc = './libensemble/tests/.coveragerc'
    with open(coveragerc, 'r') as f:
        lines = f.readlines()

    newlines = [i for i in lines if i != '    */balsam_executor.py\n']

    print('New libensemble/tests/.coveragerc: \n')
    with open(coveragerc, 'w') as f:
        for line in newlines:
            print(line)
            f.write(line)


if int(sys.version[2]) >= 6:  # Balsam only supports Python 3.6+
    install_balsam()
    move_test_balsam()
    configure_coverage()



"""
This module contains multiple generation functions for sampling a domain. All
use (and return) a random stream in ``persis_info``, given by the allocation
function.
"""
import numpy as np

__all__ = ['uniform_random_sample',
           'uniform_random_sample_with_variable_resources',
           'uniform_random_sample_obj_components',
           'latin_hypercube_sample',
           'uniform_random_sample_cancel']


def uniform_random_sample(H, persis_info, gen_specs, _):
    """
    Generates ``gen_specs['user']['gen_batch_size']`` points uniformly over the domain
    defined by ``gen_specs['user']['ub']`` and ``gen_specs['user']['lb']``.

    .. seealso::
        `test_uniform_sampling.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_uniform_sampling.py>`_ # noqa
    """
    ub = gen_specs['user']['ub']
    lb = gen_specs['user']['lb']

    n = len(lb)
    b = gen_specs['user']['gen_batch_size']

    H_o = np.zeros(b, dtype=gen_specs['out'])

    H_o['x'] = persis_info['rand_stream'].uniform(lb, ub, (b, n))

    return H_o, persis_info


def uniform_random_sample_with_variable_resources(H, persis_info, gen_specs, _):
    """
    Generates points uniformly over the domain defined by ``gen_specs['user']['ub']`` and
    ``gen_specs['user']['lb']``. Also randomly requests a different number of resource
    sets to be used in the evaluation of the generated points after the initial batch.

    .. seealso::
        `test_uniform_sampling_with_variable_resources.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_uniform_sampling_with_variable_resources.py>`_ # noqa
    """
    ub = gen_specs['user']['ub']
    lb = gen_specs['user']['lb']
    n = len(lb)

    if len(H) == 0:
        b = gen_specs['user']['initial_batch_size']

        H_o = np.zeros(b, dtype=gen_specs['out'])
        for i in range(0, b):
            # x= i*np.ones(n)
            x = persis_info['rand_stream'].uniform(lb, ub, (1, n))
            H_o['x'][i] = x
            H_o['resource_sets'][i] = 1
            H_o['priority'] = 1

    else:
        H_o = np.zeros(1, dtype=gen_specs['out'])
        # H_o['x'] = len(H)*np.ones(n)  # Can use a simple count for testing.
        H_o['x'] = persis_info['rand_stream'].uniform(lb, ub)
        H_o['resource_sets'] = persis_info['rand_stream'].randint(1, gen_specs['user']['max_resource_sets']+1)
        H_o['priority'] = 10*H_o['resource_sets']
        # print('Created sim for {} workers'.format(H_o['resource_sets']), flush=True)

    return H_o, persis_info


def uniform_random_sample_obj_components(H, persis_info, gen_specs, _):
    """
    Generates points uniformly over the domain defined by ``gen_specs['user']['ub']``
    and ``gen_specs['user']['lb']`` but requests each ``obj_component`` be evaluated
    separately.

    .. seealso::
        `test_uniform_sampling_one_residual_at_a_time.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_uniform_sampling_one_residual_at_a_time.py>`_ # noqa
    """
    ub = gen_specs['user']['ub']
    lb = gen_specs['user']['lb']

    n = len(lb)
    m = gen_specs['user']['components']
    b = gen_specs['user']['gen_batch_size']

    H_o = np.zeros(b*m, dtype=gen_specs['out'])
    for i in range(0, b):
        x = persis_info['rand_stream'].uniform(lb, ub, (1, n))
        H_o['x'][i*m:(i+1)*m, :] = np.tile(x, (m, 1))
        H_o['priority'][i*m:(i+1)*m] = persis_info['rand_stream'].uniform(0, 1, m)
        H_o['obj_component'][i*m:(i+1)*m] = np.arange(0, m)

        H_o['pt_id'][i*m:(i+1)*m] = len(H)//m+i

    return H_o, persis_info


def uniform_random_sample_cancel(H, persis_info, gen_specs, _):
    """
    Similar to uniform_random_sample but with immediate cancellation of
    selected points for testing.

    """
    ub = gen_specs['user']['ub']
    lb = gen_specs['user']['lb']

    n = len(lb)
    b = gen_specs['user']['gen_batch_size']

    H_o = np.zeros(b, dtype=gen_specs['out'])
    for i in range(b):
        if i % 10 == 0:
            H_o[i]['cancel_requested'] = True

    H_o['x'] = persis_info['rand_stream'].uniform(lb, ub, (b, n))

    return H_o, persis_info


def latin_hypercube_sample(H, persis_info, gen_specs, _):
    """
    Generates ``gen_specs['user']['gen_batch_size']`` points in a Latin
    hypercube sample over the domain defined by ``gen_specs['user']['ub']`` and
    ``gen_specs['user']['lb']``.

    .. seealso::
        `test_1d_sampling.py <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_1d_sampling.py>`_ # noqa
    """

    ub = gen_specs['user']['ub']
    lb = gen_specs['user']['lb']

    n = len(lb)
    b = gen_specs['user']['gen_batch_size']

    H_o = np.zeros(b, dtype=gen_specs['out'])

    A = lhs_sample(n, b, persis_info['rand_stream'])

    H_o['x'] = A*(ub-lb)+lb

    return H_o, persis_info


def lhs_sample(n, k, stream):

    # Generate the intervals and random values
    intervals = np.linspace(0, 1, k+1)
    rand_source = stream.uniform(0, 1, (k, n))
    rand_pts = np.zeros((k, n))
    sample = np.zeros((k, n))

    # Add a point uniformly in each interval
    a = intervals[:k]
    b = intervals[1:]
    for j in range(n):
        rand_pts[:, j] = rand_source[:, j]*(b-a) + a

    # Randomly perturb
    for j in range(n):
        sample[:, j] = rand_pts[stream.permutation(k), j]

    return sample



import numpy as np
import pprint
from libensemble.api import Ensemble
from libensemble.version import __version__
import libensemble.tests.unit_tests.setup as setup


def test_ensemble_init():
    """ Only testing attrs most likely to encounter errors """
    e = Ensemble()
    assert all([i in e.libE_specs for i in ['comms', 'mpi_comm']]), \
        "parse_args() didn't populate default values for the class instance's libE_specs"

    assert e.logger.get_level() == 20, \
        "Default log level should be 20."

    assert e._filename == __file__, \
        "Class instance's detected calling script not correctly set."


def test_from_yaml():
    """ Test that Ensemble() specs dicts resemble setup dicts """
    e = Ensemble()
    e.from_yaml('./simdir/test_example.yaml')

    sim_specs, gen_specs, exit_criteria = setup.make_criteria_and_specs_0()
    # unit_test sample specs dont (but perhaps should) contain 'user' field
    sim_specs.update({'user': {}})

    assert all([e.alloc_specs[i] is not None for i in ['user']]), \
        "class instance's alloc_specs wasn't populated."

    assert e.exit_criteria == exit_criteria, \
        "exit_criteria wasn't correctly loaded or separated from libE_specs."

    assert e.sim_specs == sim_specs, \
        "instance's sim_specs isn't equivalent to sample sim_specs"

    # can't specify np arrays in yaml - have to manually update.
    e.gen_specs['user']['ub'] = np.ones(1)
    e.gen_specs['user']['lb'] = np.zeros(1)

    assert e.gen_specs == gen_specs, \
        "instance's gen_specs isn't equivalent to sample gen_specs"


def test_str_rep():
    """ Test that Ensemble() string rep resembles setup dicts string reps """
    e = Ensemble()
    e.from_yaml('./simdir/test_example.yaml')

    sim_specs, gen_specs, exit_criteria = setup.make_criteria_and_specs_0()

    e.gen_specs['user']['ub'] = np.ones(1)
    e.gen_specs['user']['lb'] = np.zeros(1)

    repd = str(e).split()

    unmatched_strings = []

    for spec in [sim_specs, gen_specs, exit_criteria]:
        for i in pprint.pformat(spec).split():
            if i in repd:
                continue
            else:
                unmatched_strings.append(i)

    # Only possible unmatches should be addresses?
    assert all(['0x' in item for item in unmatched_strings]), \
        "String representation components didn\'t match expected."

    assert __version__ in repd, \
        "libEnsemble version not detected in string representation."


def test_bad_func_loads():
    """ Test that Ensemble() raises expected errors (with warnings) on incorrect imports """

    yaml_errors = {'./simdir/test_example_badfuncs_attribute.yaml': AttributeError,
                   './simdir/test_example_badfuncs_notfound.yaml': ModuleNotFoundError}

    for f in yaml_errors:
        e = Ensemble()
        flag = 1
        try:
            e.from_yaml(f)
        except yaml_errors[f]:
            flag = 0
        assert flag == 0


if __name__ == '__main__':
    test_ensemble_init()
    test_from_yaml()
    test_str_rep()
    test_bad_func_loads()



import numpy as np
import subprocess
import os

from libensemble.message_numbers import STOP_TAG, PERSIS_STOP, FINISHED_PERSISTENT_GEN_TAG, EVAL_GEN_TAG
from libensemble.tools.persistent_support import PersistentSupport


def build_H0(x_f_pairs, gen_specs, noise_h_mat):

    U = gen_specs['user']
    x0 = U['x0']

    # This function constructs H0 to contain points to be sent back to the
    # manager to be evaluated

    n = len(x0)
    E = np.eye(n)
    nf = U['nf']

    H0 = np.zeros(len(x_f_pairs)*nf, dtype=gen_specs['out'])
    ind = 0
    for i, j in x_f_pairs:
        for k in range(nf+1):
            if k != nf//2:
                H0['x'][ind] = x0 + (k-nf/2)*noise_h_mat[i, j]*E[i]
                H0['x_ind'][ind] = i
                H0['f_ind'][ind] = j
                H0['n_ind'][ind] = k
                ind += 1

    return H0


def fd_param_finder(H, persis_info, gen_specs, libE_info):
    """
    This generation function loops through a set of suitable finite difference
    parameters for a mapping F from R^n to R^m.

    .. seealso::
        `test_persistent_fd_param_finder.py` <https://github.com/Libensemble/libensemble/blob/develop/libensemble/tests/regression_tests/test_persistent_fd_param_finder.py>`_ # noqa
    """
    U = gen_specs['user']

    p = U['p']
    x0 = U['x0']
    nf = U['nf']
    noise_h_mat = U['noise_h_mat']
    inform = np.zeros_like(noise_h_mat)
    Fnoise = np.zeros_like(noise_h_mat)
    maxnoiseits = U['maxnoiseits']
    ps = PersistentSupport(libE_info, EVAL_GEN_TAG)

    n = len(x0)
    Fhist0 = np.zeros((n, p, nf+1))
    tag = None

    # # Request evaluations of the base point x0 at all p f_inds
    # H0 = np.zeros(p, dtype=gen_specs['out'])
    # for j in range(p):
    #     H0['x'][j] = x0
    #     H0['x_ind'][j] = -1  # Marking these to know they are the basepoint
    #     H0['f_ind'][j] = j
    #     H0['n_ind'][j] = nf/2
    # tag, Work, calc_in = sendrecv_mgr_worker_msg(comm, H0)
    for i in range(n):
        for j in range(p):
            # Fhist0[i,j,nf//2] = calc_in['f_val'][calc_in['f_ind']==j]
            Fhist0[i, j, nf//2] = U['f0'][j]

    x_f_pairs = np.array(np.meshgrid(range(n), range(p))).T.reshape(-1, n)
    H0 = build_H0(x_f_pairs, gen_specs, noise_h_mat)

    iters = np.ones_like(noise_h_mat)

    tag, Work, calc_in = ps.send_recv(H0)

    # import matlab.engine
    # eng = matlab.engine.start_matlab()

    # Send nf points for each (x_ind, f_ind) pair
    while tag not in [STOP_TAG, PERSIS_STOP]:
        x_f_pairs = np.unique(calc_in[['x_ind', 'f_ind']])
        x_f_pairs_new = []

        # Update Fhist0
        for i, j in x_f_pairs:
            for k in range(nf+1):
                if k != nf/2:
                    logical_conds = (calc_in['x_ind'] == i, calc_in['f_ind'] == j, calc_in['n_ind'] == k)
                    Fhist0[i, j, k] = calc_in['f_val'][np.logical_and.reduce(logical_conds)]

            # Compute noise for (i,j):
            # [Fnoise(i,j),~,inform(i,j)] = ECnoise(nf-1,Fhist0(i,j,2:nf));
            # t = eng.ECnoise(nf+1,matlab.double(Fhist0[i,j,:nf+1]),nargout=3)
            # # Optional: check to see what would get with 2 fewer evals (requires nf>=4):
            # [Fnoise2(i,j),~,inform2(i,j)] = ECnoise(nf-1,Fhist0(i,j,2:nf));

            # cmd = ["/home/jlarson/software/MATLAB/R2019a/bin/matlab", "-batch",
            cmd = ["octave", "--no-window-system", "--eval",
                   "F=[" + " ".join(["{:18.18f}".format(x) for x in Fhist0[i, j, :nf+1]]) + "];"
                   "nf=" + str(nf) + "';"
                   "[fnoise,~,inform] = ECnoise(nf+1,F);"
                   "dlmwrite('fnoise.out', fnoise, 'delimiter', ' ', 'precision', 16);"
                   "dlmwrite('inform.out', inform, 'delimiter', ' ', 'precision', 16);"
                   "exit"
                   ]

            p = subprocess.call(cmd, shell=False, stdout=subprocess.DEVNULL)

            inform[i, j] = np.loadtxt('inform.out')

            if inform[i, j] >= 2:
                # Mark as needing more points for this noise_h_mat value
                if iters[i, j] < maxnoiseits:
                    iters[i, j] += 1
                    x_f_pairs_new.append((i, j))

                    if inform[i, j] == 3:
                        noise_h_mat[i, j] = noise_h_mat[i, j]/100
                    else:
                        noise_h_mat[i, j] = noise_h_mat[i, j]*100
            else:
                # We have successfully identified the Fnoise
                Fnoise[i, j] = np.loadtxt('fnoise.out')

            os.remove('inform.out')
            os.remove('fnoise.out')

        if np.all(inform == 1):
            break

        H0 = build_H0(x_f_pairs_new, gen_specs, noise_h_mat)
        tag, Work, calc_in = ps.send_recv(H0)

    persis_info['Fnoise'] = Fnoise
    return H0, persis_info, FINISHED_PERSISTENT_GEN_TAG



import os
import subprocess
import sys
import time
import signal


# Does not kill, even on laptop
# kill 1
def kill_task_1(process):
    process.kill()
    process.wait()


# kill 2 - with  preexec_fn=os.setsid in subprocess
def kill_task_2(process):
    os.killpg(os.getpgid(process.pid), signal.SIGKILL)
    process.wait()


if len(sys.argv) != 4:
    raise Exception("Usage: python killtest.py <kill_type> <num_nodes> <num_procs_per_node>")

# user_code = "./burn_time.x"
user_code = "./sleep_and_print.x"

# sys.argv[0] is python exe.
kill_type = int(sys.argv[1])  # 1, 2
num_nodes = int(sys.argv[2])
num_procs_per_node = int(sys.argv[3])
num_procs = num_nodes * num_procs_per_node

print("Running Kill test with program", user_code)
print("Kill type: {}   num_nodes: {}   procs_per_node: {}".format(kill_type, num_nodes, num_procs_per_node))


# Create common components of submit line (currently all of it)

# Am I in an aprun environment
launcher = 'mpich'  # Includes mpich based - eg. intelmpi
try:
    subprocess.check_call(['aprun', '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
except OSError:
    launcher = 'mpich'
else:
    launcher = 'aprun'

if launcher == 'mpich':
    mpicmd_launcher = 'mpirun'
    mpicmd_numprocs = '-np'
    mpicmd_ppn = '-ppn'
elif launcher == 'aprun':
    mpicmd_launcher = 'aprun'
    mpicmd_numprocs = '-n'
    mpicmd_ppn = '-N'

# As runline common to tasks currently - construct here.
runline = []                             # E.g: 2 nodes run
runline.append(mpicmd_launcher)          # mpirun
runline.append(mpicmd_numprocs)          # mpirun -np
runline.append(str(num_procs))           # mpirun -np 8
runline.append(mpicmd_ppn)               # mpirun -np 8 --ppn
runline.append(str(num_procs_per_node))  # mpirun -np 8 --ppn 4
runline.append(user_code)                # mpirun -np 8 --ppn 4 ./burn_time.x


# print("Running killtest.py with task size {} procs".format(num_procs))
total_start_time = time.time()

for run_num in range(2):
    time.sleep(4)  # Show gap where none should be running
    stdout = "out_" + str(run_num) + ".txt"
    # runline = ['mpirun', '-np', str(num_procs), user_code]
    print('---------------------------------------------------------------')
    print('\nRun num: {}   Runline: {}\n'.format(run_num, " ".join(runline)))

    if kill_type == 1:
        process = subprocess.Popen(runline, cwd='./', stdout=open(stdout, 'w'), shell=False)  # with kill 1
    elif kill_type == 2:
        process = subprocess.Popen(runline, cwd='./', stdout=open(stdout, 'w'), shell=False,
                                   preexec_fn=os.setsid)  # kill 2
    else:
        raise Exception("kill_type not recognized")

    time_limit = 4
    start_time = time.time()
    finished = False
    state = "Not started"
    while(not finished):

        time.sleep(2)
        poll = process.poll()
        if poll is None:
            state = 'RUNNING'
            print('Running....')

        else:
            finished = True
            if process.returncode == 0:
                state = 'PASSED'
            else:
                state = 'FAILED'

        if(time.time() - start_time > time_limit):
            print('Killing task', run_num)
            # kill_task(process, user_code)

            if kill_type == 1:
                kill_task_1(process)
            elif kill_type == 2:
                kill_task_2(process)
            state = 'KILLED'
            finished = True

    # Assert task killed
    assert state == 'KILLED', "Task not registering as killed. State is: " + state

    # Checking if processes still running and producing output
    grace_period = 1    # Seconds after kill when first read last line
    recheck_period = 2  # Recheck last line after this many seconds
    num_rechecks = 2    # Number of times to check for new output

    time.sleep(grace_period)  # Give chance to kill

    # Test if task is still producing output
    with open(stdout, 'rb') as fh:
        line_on_kill = fh.readlines()[-1].decode().rstrip()
    print("Last line after task kill:  {}".format(line_on_kill))

    if 'has finished' in line_on_kill:
        raise Exception('Task may have already finished - test invalid')

    for recheck in range(1, num_rechecks+1):
        time.sleep(recheck_period)
        with open(stdout, 'rb') as fh:
            lastline = fh.readlines()[-1].decode().rstrip()
        print("Last line after {} seconds: {}".format(recheck_period*recheck, lastline))

        if lastline != line_on_kill:
            print("Task {} still producing output".format(run_num))
            # print("Last line check 1:", line_on_kill)
            # print("Last line check 2:", lastline)
            assert 0

total_end_time = time.time()
total_time = total_end_time - total_start_time
print("\nTask kill test completed in {} seconds\n".format(total_time))



import subprocess
import os
import time
import libensemble
from libensemble.tests.regression_tests.common import modify_Balsam_worker, modify_Balsam_JobEnv

# TESTSUITE_COMMS: local
# TESTSUITE_NPROCS: 3
# TESTSUITE_EXTRA: true

# This test is NOT submitted as a job to Balsam. script_test_balsam_hworld.py is
#   the executable submitted to Balsam as a job. This test executes that job
#   through the 'runstr' line in run_Balsam_job()


def run_Balsam_job():
    runstr = 'balsam launcher --consume-all --job-mode=mpi --num-transition-threads=1'
    print('Executing Balsam job with command: {}'.format(runstr))
    subprocess.Popen(runstr.split())


def build_simfunc():
    buildstring = 'mpicc -o my_simtask.x ../unit_tests/simdir/my_simtask.c'
    subprocess.check_call(buildstring.split())


def wait_for_job_dir(basedb):
    sleeptime = 0
    limit = 15

    # Stop sleeping once database directory detected
    print('Waiting for Balsam Database directory.')
    while sleeptime < limit:
        if os.path.isdir(basedb):
            break
        time.sleep(1)
        sleeptime += 1

    assert sleeptime < limit, \
        "Balsam Database directory not created within {} seconds.".format(limit)

    # Stop sleeping once job directory detected within database directory
    print('Waiting for Job Directory {}'.format(sleeptime))
    while sleeptime < limit:
        if len(os.listdir(basedb)) > 0:
            break
        print(sleeptime, end=" ", flush=True)
        time.sleep(1)
        sleeptime += 1

    assert sleeptime < limit, \
        "Balsam Job directory not created within {} seconds.".format(limit)

    # Assumes database dir was empty, now contains single job dir
    jobdir = os.path.join(basedb, os.listdir(basedb)[0])
    return jobdir


def wait_for_job_output(jobdir):
    sleeptime = 0
    limit = 60

    output = os.path.join(jobdir, 'job_script_test_balsam_hworld.out')
    print('Checking for Balsam output file: {}'.format(output))

    while sleeptime < limit:
        if os.path.isfile(output):
            break
        print(sleeptime, end=" ", flush=True)
        print(os.listdir(jobdir), flush=True)
        time.sleep(1)
        sleeptime += 1

    assert sleeptime < limit, \
        "Balsam output file not created within {} seconds.".format(limit)

    return output


def print_job_output(outscript):
    sleeptime = 0
    limit = 80

    print('Blank output file found. Waiting for expected complete Balsam Job Output.')
    succeed_line = 'Received:  [34 34 31 31 34 34 32 32 33 33  0  0]\n'

    lastposition = 0

    while sleeptime < limit:
        with open(outscript, 'r') as f:
            f.seek(lastposition)
            new = f.read()
            lastposition = f.tell()

        if len(new) > 0:
            print(new, flush=True)
        else:
            print(sleeptime, end=" ", flush=True)

        if succeed_line in new:
            print('Success. Received task statuses match expected.')
            break

        time.sleep(1)
        sleeptime += 1

    assert sleeptime < limit, \
        "Expected Balsam Job output-file contents not detected after {} seconds.".format(limit)


def move_job_coverage(jobdir):
    # Move coverage files from Balsam DB to ./regression_tests (for concatenation)
    print('Moving job coverage results.')
    here = os.getcwd()
    covname = '.cov_reg_out.'

    assert any([file.startswith(covname) for file in os.listdir(jobdir)]), \
        "Coverage results not detected in Balsam Job directory."

    for file in os.listdir(jobdir):
        if file.startswith(covname):
            balsam_cov = os.path.join(jobdir, file)
            here_cov = os.path.join(here, file)
            os.rename(balsam_cov, here_cov)


if __name__ == '__main__':

    # Used by Balsam Coverage config file. Dont evaluate Balsam data dir
    libepath = os.path.dirname(libensemble.__file__)
    os.environ['LIBE_PATH'] = libepath
    # os.environ['BALSAM_DB_PATH'] = '~/test-balsam'

    basedb = os.environ['HOME'] + '/test-balsam/data/libe_test-balsam'

    subprocess.run('./scripts_used_by_reg_tests/configure-balsam-test.sh'.split())

    if not os.path.isfile('./my_simtask.x'):
        build_simfunc()

    modify_Balsam_worker()
    modify_Balsam_JobEnv()
    run_Balsam_job()

    jobdir = wait_for_job_dir(basedb)
    output = wait_for_job_output(jobdir)
    print_job_output(output)
    move_job_coverage(jobdir)

    print('Test complete.')
    subprocess.run('./scripts_used_by_reg_tests/cleanup-balsam-test.sh'.split())



import os
import numpy as np
from libensemble.tools.fields_keys import libE_fields, allowed_gen_spec_keys, \
    allowed_sim_spec_keys, allowed_alloc_spec_keys, allowed_libE_spec_keys


def _check_consistent_field(name, field0, field1):
    """Checks that new field (field1) is compatible with an old field (field0)."""
    assert field0.ndim == field1.ndim, \
        "H0 and H have different ndim for field {}".format(name)
    assert (np.all(np.array(field1.shape) >= np.array(field0.shape))), \
        "H too small to receive all components of H0 in field {}".format(name)


def check_libE_specs(libE_specs, serial_check=False):
    assert isinstance(libE_specs, dict), "libE_specs must be a dictionary"

    comms_type = libE_specs.get('comms', 'mpi')
    if comms_type in ['mpi']:
        if not serial_check:
            assert libE_specs['mpi_comm'].Get_size() > 1, "Manager only - must be at least one worker (2 MPI tasks)"
    elif comms_type in ['local']:
        assert libE_specs['nworkers'] >= 1, "Must specify at least one worker"
    elif comms_type in ['tcp']:
        # TODO, differentiate and test SSH/Client
        assert libE_specs['nworkers'] >= 1, "Must specify at least one worker"

    for k in libE_specs.keys():
        assert k in allowed_libE_spec_keys,\
            "Key %s is not allowed in libE_specs. Supported keys are: %s " % (k, allowed_libE_spec_keys)

        if k in ['ensemble_copy_back', 'use_worker_dirs', 'sim_dirs_make', 'gen_dirs_make']:
            assert isinstance(libE_specs[k], bool), "Value for libE_specs['{}'] must be Boolean".format(k)

        if k in ['sim_input_dir', 'gen_input_dir']:
            assert isinstance(libE_specs[k], str), \
                "Value for libE_specs['{}'] must be a single path-like string".format(k)
            assert os.path.exists(libE_specs[k]), \
                "libE_specs['{}'] does not refer to an existing path.".format(k)

        if k == 'ensemble_dir_path':
            assert isinstance(libE_specs[k], str), \
                "Value for libE_specs['{}'] must be a single path-like string".format(k)

        if k in ['sim_dir_copy_files', 'sim_dir_symlink_files', 'gen_dir_copy_files', 'gen_dir_symlink_files']:
            assert isinstance(libE_specs[k], list), \
                "Value for libE_specs['{}'] must be a list of path-like strings".format(k)
            for j in libE_specs[k]:
                assert os.path.exists(j), \
                    "'{}' in libE_specs['{}'] does not refer to an existing path.".format(j, k)


def check_alloc_specs(alloc_specs):
    assert isinstance(alloc_specs, dict), "alloc_specs must be a dictionary"

    assert alloc_specs['alloc_f'], "Allocation function must be specified"

    for k in alloc_specs.keys():
        assert k in allowed_alloc_spec_keys,\
            "Key %s is not allowed in alloc_specs. Supported keys are: %s " % (k, allowed_alloc_spec_keys)


def check_sim_specs(sim_specs):
    assert isinstance(sim_specs, dict), "sim_specs must be a dictionary"

    assert any([term_field in sim_specs for term_field in ['sim_f', 'in', 'persis_in', 'out']]), \
        "sim_specs must contain 'sim_f', 'in', 'out'"

    assert all(isinstance(i, str) for i in sim_specs['in']), \
        "Entries in sim_specs['in'] must be strings. Also can't be lists or tuples of strings."

    assert len(sim_specs['out']), "sim_specs must have 'out' entries"

    assert isinstance(sim_specs['in'], list), "'in' field must exist and be a list of field names"

    for k in sim_specs.keys():
        assert k in allowed_sim_spec_keys,\
            "Key %s is not allowed in sim_specs. Supported keys are: %s " % (k, allowed_sim_spec_keys)


def check_gen_specs(gen_specs):
    assert isinstance(gen_specs, dict), "gen_specs must be a dictionary"

    assert not bool(gen_specs) or len(gen_specs['out']), "gen_specs must have 'out' entries"

    if 'in' in gen_specs:
        assert all(isinstance(i, str) for i in gen_specs['in']), \
            "Entries in gen_specs['in'] must be strings. Also can't be lists or tuples of strings."

    for k in gen_specs.keys():
        assert k in allowed_gen_spec_keys,\
            "Key %s is not allowed in gen_specs. Supported keys are: %s " % (k, allowed_gen_spec_keys)


def check_exit_criteria(exit_criteria, sim_specs, gen_specs):
    assert isinstance(exit_criteria, dict), "exit_criteria must be a dictionary"

    assert len(exit_criteria) > 0, "Must have some exit criterion"

    # Ensure termination criteria are valid
    valid_term_fields = ['sim_max', 'gen_max',
                         'elapsed_wallclock_time', 'stop_val']
    assert all([term_field in valid_term_fields for term_field in exit_criteria]), \
        "Valid termination options: " + str(valid_term_fields)

    # Make sure stop-values match parameters in gen_specs or sim_specs
    if 'stop_val' in exit_criteria:
        stop_name = exit_criteria['stop_val'][0]
        sim_out_names = [e[0] for e in sim_specs['out']]
        gen_out_names = [e[0] for e in gen_specs['out']]
        assert stop_name in sim_out_names + gen_out_names, \
            "Can't stop on {} if it's not in a sim/gen output".format(stop_name)


def check_H(H0, sim_specs, alloc_specs, gen_specs):
    if len(H0):
        # Set up dummy history to see if it agrees with H0

        # Combines all 'out' fields (if they exist) in sim_specs, gen_specs, or alloc_specs
        specs = [sim_specs, alloc_specs, gen_specs]
        dtype_list = list(set(libE_fields + sum([k.get('out', []) for k in specs if k], [])))
        Dummy_H = np.zeros(1 + len(H0), dtype=dtype_list)

        fields = H0.dtype.names

        # Prior history must contain the fields in new history
        assert set(fields).issubset(set(Dummy_H.dtype.names)), \
            "H0 contains fields {} not in the History.".\
            format(set(fields).difference(set(Dummy_H.dtype.names)))

        # Prior history cannot contain unreturned points
        # assert 'returned' not in fields or np.all(H0['returned']), \
        #     "H0 contains unreturned points."

        # Fail if prior history contains unreturned points (or returned but not given).
        assert('returned' not in fields or np.all(H0['given'] == H0['returned'])), \
            'H0 contains unreturned or invalid points'

        # # Fail if points in prior history don't have a sim_id.
        # assert('sim_id' in fields), 'Points in H0 must have sim_ids'

        # Check dimensional compatibility of fields
        for field in fields:
            _check_consistent_field(field, H0[field], Dummy_H[field])


def check_inputs(libE_specs=None, alloc_specs=None, sim_specs=None,
                 gen_specs=None, exit_criteria=None, H0=None, serial_check=False):
    """
    Checks whether the libEnsemble arguments are of the correct data type and
    contain sufficient information to perform a run. There is no return value.
    An exception is raised if any of the checks fail.

    .. code-block:: python

        from libensemble.tools import check_inputs
        check_inputs(sim_specs=my_sim_specs, gen_specs=my_gen_specs, exit_criteria=ec)

    Parameters
    ----------

    libE_specs, alloc_specs, sim_specs, gen_specs, exit_criteria: :obj:`dict`, optional

        libEnsemble data structures

    H0: :obj:`numpy structured array`, optional

        A previous libEnsemble history to be prepended to the history in the
        current libEnsemble run
        :doc:`(example)<data_structures/history_array>`

    serial_check : :obj:`boolean`

        If true, assumes running a serial check. This means, for example,
        the details of current MPI communicator are not checked (can be
        run with libE_specs{'mpi_comm': 'mpi'} without running through mpiexec.

    """
    out_names = [e[0] for e in libE_fields]
    if H0 is not None and H0.dtype.names is not None:
        out_names += list(H0.dtype.names)
    if sim_specs is not None:
        out_names += [e[0] for e in sim_specs.get('out', [])]
    if gen_specs is not None:
        out_names += [e[0] for e in gen_specs.get('out', [])]
    if alloc_specs is not None:
        out_names += [e[0] for e in alloc_specs.get('out', [])]

    # Detailed checking based on Required Keys in docs for each specs
    if libE_specs is not None:
        for name in libE_specs.get('final_fields', []):
            assert name in out_names, \
                name + " in libE_specs['fields_keys'] is not in sim_specs['out'], "\
                "gen_specs['out'], alloc_specs['out'], H0, or libE_fields."
        check_libE_specs(libE_specs, serial_check)

    if alloc_specs is not None:
        for name in alloc_specs.get('in', []):
            assert name in out_names, \
                name + " in alloc_specs['in'] is not in sim_specs['out'], "\
                "gen_specs['out'], alloc_specs['out'], H0, or libE_fields."

        check_alloc_specs(alloc_specs)

    if sim_specs is not None:
        for name in sim_specs.get('in', []):
            assert name in out_names, \
                name + " in sim_specs['in'] is not in sim_specs['out'], "\
                "gen_specs['out'], alloc_specs['out'], H0, or libE_fields."

        check_sim_specs(sim_specs)

    if gen_specs is not None:
        for name in gen_specs.get('in', []):
            assert name in out_names, \
                name + " in gen_specs['in'] is not in sim_specs['out'], "\
                "gen_specs['out'], alloc_specs['out'], H0, or libE_fields."

        check_gen_specs(gen_specs)

    if exit_criteria is not None:
        assert sim_specs is not None and gen_specs is not None, \
            "Can't check exit_criteria without sim_specs and gen_specs"
        check_exit_criteria(exit_criteria, sim_specs, gen_specs)

    if H0 is not None:
        assert sim_specs is not None and alloc_specs is not None and gen_specs is not None, \
            "Can't check H0 without sim_specs, alloc_specs, gen_specs"
        check_H(H0, sim_specs, alloc_specs, gen_specs)



"""
Runs libEnsemble with APOSMM+DFOLS on the chwirut least-squares problem.
All 214 residual calculations for a given point are performed as a single
simulation evaluation.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_persistent_aposmm_dfols.py
   python3 test_persistent_aposmm_dfols.py --nworkers 3 --comms local
   python3 test_persistent_aposmm_dfols.py --nworkers 3 --comms tcp

When running with the above commands, the number of concurrent evaluations of
the objective function will be 2, as one of the three workers will be the
persistent generator.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: local mpi tcp
# TESTSUITE_NPROCS: 4
# TESTSUITE_EXTRA: true

import sys
import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.chwirut1 import chwirut_eval as sim_f

import libensemble.gen_funcs

libensemble.gen_funcs.rc.aposmm_optimizers = 'dfols'

from libensemble.gen_funcs.persistent_aposmm import aposmm as gen_f
from libensemble.alloc_funcs.persistent_aposmm_alloc import persistent_aposmm_alloc as alloc_f
from libensemble.tools import parse_args, save_libE_output, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()

if nworkers < 2:
    sys.exit("Cannot run with a persistent worker if only one worker -- aborting...")

# Declare the run parameters/functions
m = 214
n = 3
budget = 10

sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float), ('fvec', float, m)],
    'user': {
        'combine_component_func': lambda x: np.sum(np.power(x, 2)),
    },
}

gen_out = [('x', float, n), ('x_on_cube', float, n), ('sim_id', int), ('local_min', bool), ('local_pt', bool)]

# lb tries to avoid x[1]=-x[2], which results in division by zero in chwirut.
gen_specs = {
    'gen_f': gen_f,
    'persis_in': ['f', 'fvec'] + [n[0] for n in gen_out],
    'out': gen_out,
    'user': {
        'initial_sample_size': 100,
        'localopt_method': 'dfols',
        'components': m,
        'dfols_kwargs': {
            'do_logging': False,
            'rhoend': 1e-5,
            'user_params': {
                'model.abs_tol': 1e-10,
                'model.rel_tol': 1e-4,
            },
        },
        'lb': (-2 - np.pi / 10) * np.ones(n),
        'ub': 2 * np.ones(n),
    },
}

alloc_specs = {'alloc_f': alloc_f}

persis_info = add_unique_random_streams({}, nworkers + 1)

# Tell libEnsemble when to stop (stop_val key must be in H)
exit_criteria = {
    'sim_max': 1000,
    'elapsed_wallclock_time': 100,
    'stop_val': ('f', 3000),
}
# end_exit_criteria_rst_tag

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)

if is_manager:
    assert persis_info[1].get('run_order'), "Run_order should have been given back"
    assert flag == 0
    assert np.min(H['f'][H['returned']]) <= 3000, "Didn't find a value below 3000"

    save_libE_output(H, persis_info, __file__, nworkers)

    # # Calculating the Jacobian at local_minima (though this information was not used by DFO-LS)
    from libensemble.sim_funcs.chwirut1 import EvaluateFunction, EvaluateJacobian

    for i in np.where(H['local_min'])[0]:
        F = EvaluateFunction(H['x'][i])
        J = EvaluateJacobian(H['x'][i])
    #     u = gen_specs['user']['ub']-H['x'][i]
    #     l = H['x'][i]-gen_specs['user']['lb']
    #     if np.any(u <= 1e-7) or np.any(l <= 1e-7):
    #         grad = -2*np.dot(J.T,F)
    #         assert np.all(grad[u<=1e-7] >= 0)
    #         assert np.all(grad[l<=1e-7] <= 0)

    #         if not np.all(grad[np.logical_and(u>=1e-7,l>=1e-7)] <= 1e-5):
    #             import ipdb; ipdb.set_trace()
    #     else:
    #         d = np.linalg.solve(np.dot(J.T,J),np.dot(J.T,F))
    #         assert np.linalg.norm(d) <= 1e-5



#!/usr/bin/env python
import os
import numpy as np
from tutorial_forces_simf import run_forces  # Sim func from current dir

from libensemble.libE import libE
from libensemble.gen_funcs.sampling import uniform_random_sample
from libensemble.tools import parse_args, add_unique_random_streams
from libensemble.executors.mpi_executor import MPIExecutor

nworkers, is_manager, libE_specs, _ = parse_args()  # Convenience function

# Create executor and register sim to it
exctr = MPIExecutor()

# Register simulation executable with executor
sim_app = os.path.join(os.getcwd(), 'forces.x')
exctr.register_app(full_path=sim_app, app_name='forces')

# State the sim_f, its arguments, output, and parameters (and their sizes)
sim_specs = {'sim_f': run_forces,         # sim_f, imported above
             'in': ['x'],                 # Name of input for sim_f
             'out': [('energy', float)],  # Name, type of output from sim_f
             'user': {'simdir_basename': 'forces',  # User parameters for the sim_f
                      'keys': ['seed'],
                      'cores': 2,
                      'sim_particles': 1e3,
                      'sim_timesteps': 5,
                      'sim_kill_minutes': 10.0,
                      'particle_variance': 0.2,
                      'kill_rate': 0.5}
             }

# State the gen_f, its arguments, output, and necessary parameters.
gen_specs = {'gen_f': uniform_random_sample,  # Generator function
             'in': ['sim_id'],                # Generator input
             'out': [('x', float, (1,))],     # Name, type and size of data from gen_f
             'user': {'lb': np.array([0]),             # User parameters for the gen_f
                      'ub': np.array([32767]),
                      'gen_batch_size': 1000,
                      'batch_mode': True,
                      'num_active_gens': 1,
                      }
             }

libE_specs['save_every_k_gens'] = 1000  # Save every K steps
libE_specs['sim_dirs_make'] = True

exit_criteria = {'sim_max': 8}

persis_info = add_unique_random_streams({}, nworkers + 1)

H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria,
                            persis_info=persis_info, libE_specs=libE_specs)



"""
This module stores environment variables for use in resource detection
"""

import os
import re
import logging
from collections import OrderedDict

logger = logging.getLogger(__name__)


class EnvResources:
    """Stores environment variables to query for system resource information

    **Class Attributes:**

    :cvar string default_nodelist_env_slurm: Default SLRUM nodelist environment variable
    :cvar string default_nodelist_env_cobalt: Default Cobal nodelist environment variable
    :cvar string default_nodelist_env_lsf: Default LSF nodelist environment variable
    :cvar string default_nodelist_env_lsf_shortform: Default LSF short-form nodelist environment variable

    **Object Attributes:**

    These are set on initialization.

    :ivar dict nodelists: Environment variable names to query for nodelists by scheduler
    :ivar dict ndlist_funcs: Functions to extract nodelists from environment by scheduler
    """

    default_nodelist_env_slurm = 'SLURM_NODELIST'
    default_nodelist_env_cobalt = 'COBALT_PARTNAME'
    default_nodelist_env_lsf = 'LSB_HOSTS'
    default_nodelist_env_lsf_shortform = 'LSB_MCPU_HOSTS'

    def __init__(self,
                 nodelist_env_slurm=None,
                 nodelist_env_cobalt=None,
                 nodelist_env_lsf=None,
                 nodelist_env_lsf_shortform=None):

        """Initializes a new EnvResources instance

        Determines the environment variables to query for resource
        information. These are either provided or given defaults.

        Parameters
        ----------

        nodelist_env_slurm: String, optional
            The environment variable giving a node list in Slurm format (Default: uses SLURM_NODELIST).
            Note: This is queried only if a node_list file is not provided.

        nodelist_env_cobalt: String, optional
            The environment variable giving a node list in Cobalt format (Default: uses COBALT_PARTNAME).
            Note: This is queried only if a node_list file is not provided and disable_resource_manager=True.

        nodelist_env_lsf: String, optional
            The environment variable giving a node list in LSF format (Default: uses LSB_HOSTS).
            Note: This is queried only if a node_list file is not provided.

        nodelist_env_lsf_shortform: String, optional
            The environment variable giving a node list in LSF short-form format (Default: uses LSB_MCPU_HOSTS).
            Note: This is queried only if a node_list file is not provided.
        """

        self.scheduler = None
        self.nodelists = {}
        self.nodelists['Slurm'] = nodelist_env_slurm or EnvResources.default_nodelist_env_slurm
        self.nodelists['Cobalt'] = nodelist_env_cobalt or EnvResources.default_nodelist_env_cobalt
        self.nodelists['LSF'] = nodelist_env_lsf or EnvResources.default_nodelist_env_lsf
        self.nodelists['LSF_shortform'] = nodelist_env_lsf_shortform or EnvResources.default_nodelist_env_lsf_shortform

        self.ndlist_funcs = {}
        self.ndlist_funcs['Slurm'] = EnvResources.get_slurm_nodelist
        self.ndlist_funcs['Cobalt'] = EnvResources.get_cobalt_nodelist
        self.ndlist_funcs['LSF'] = EnvResources.get_lsf_nodelist
        self.ndlist_funcs['LSF_shortform'] = EnvResources.get_lsf_nodelist_frm_shortform

        for env, env_var in self.nodelists.items():
            if os.environ.get(env_var):
                self.scheduler = env
                break

    def get_nodelist(self):
        """Returns nodelist from environment or an empty list"""
        if self.scheduler:
            env = self.scheduler
            env_var = self.nodelists[env]
            logger.debug("{} env found - getting nodelist from {}".format(env, env_var))
            get_list_func = self.ndlist_funcs[env]
            global_nodelist = get_list_func(env_var)
            return global_nodelist
        return []

    @staticmethod
    def abbrev_nodenames(node_list, prefix=None):
        """Returns nodelist with only string up to first dot"""
        newlist = [s.split(".", 1)[0] for s in node_list]
        return newlist

    @staticmethod
    def cobalt_abbrev_nodenames(node_list, prefix='nid'):
        """Returns nodelist with prefix and leading zeros stripped"""
        newlist = [s.lstrip(prefix) for s in node_list]
        newlist = [s.lstrip('0') for s in newlist]
        return newlist

    def shortnames(self, node_list):
        """Returns nodelist with entries in abbreviated form"""
        if self.scheduler == 'Cobalt':
            return EnvResources.cobalt_abbrev_nodenames(node_list)
        elif self.scheduler is not None:
            return EnvResources.abbrev_nodenames(node_list)
        return node_list

    @staticmethod
    def _range_split(s):
        """Splits ID range string"""
        ab = s.split("-", 1)
        nnum_len = len(ab[0])
        a = int(ab[0])
        b = int(ab[-1])
        if a > b:
            a, b = b, a
        b = b + 1
        return a, b, nnum_len

    @staticmethod
    def _noderange_append(prefix, nidstr, suffix):
        """Formats and appends nodes to overall nodelist"""
        nidlst = []
        for nidgroup in nidstr.split(','):
            a, b, nnum_len = EnvResources._range_split(nidgroup)
            for nid in range(a, b):
                nidlst.append(prefix + str(nid).zfill(nnum_len) + suffix)
        return nidlst

    @staticmethod
    def get_slurm_nodelist(node_list_env):
        """Gets global libEnsemble nodelist from the Slurm environment"""
        fullstr = os.environ[node_list_env]
        if not fullstr:
            return []
        # Split at commas outside of square brackets
        r = re.compile(r'(?:[^,\[]|\[[^\]]*\])+')
        part_splitstr = r.findall(fullstr)
        nidlst = []
        for i in range(len(part_splitstr)):
            part = part_splitstr[i]
            splitstr = part.split('[', 1)
            if len(splitstr) == 1:
                nidlst.append(splitstr[0])
            else:
                prefix = splitstr[0]
                remainder = splitstr[1]
                splitstr = remainder.split(']', 1)
                nidstr = splitstr[0]
                suffix = splitstr[1]
                nidlst.extend(EnvResources._noderange_append(prefix, nidstr, suffix))
        return sorted(nidlst)

    @staticmethod
    def get_cobalt_nodelist(node_list_env):
        """Gets global libEnsemble nodelist from the Cobalt environment"""
        nidlst = []
        nidstr = os.environ[node_list_env]
        if not nidstr:
            return []
        for nidgroup in nidstr.split(','):
            a, b, _ = EnvResources._range_split(nidgroup)
            for nid in range(a, b):
                nidlst.append(str(nid))
        return sorted(nidlst, key=int)

    @staticmethod
    def get_lsf_nodelist(node_list_env):
        """Gets global libEnsemble nodelist from the LSF environment"""
        full_list = os.environ[node_list_env]
        entries = full_list.split()
        # unique_entries = list(set(entries)) # This will not retain order
        unique_entries = list(OrderedDict.fromkeys(entries))
        nodes = [n for n in unique_entries if 'batch' not in n]
        return nodes

    @staticmethod
    def get_lsf_nodelist_frm_shortform(node_list_env):
        """Gets global libEnsemble nodelist from the LSF environment from short-form version"""
        full_list = os.environ[node_list_env]
        entries = full_list.split()
        iter_list = iter(entries)
        zipped_list = list(zip(iter_list, iter_list))
        nodes_with_count = [n for n in zipped_list if 'batch' not in n[0]]
        nodes = [n[0] for n in nodes_with_count]
        return nodes



"""Standalone comms test

Testing for message errors and correctness of data
"""
from mpi4py import MPI
import numpy as np
import time

comm = MPI.COMM_WORLD
num_workers = MPI.COMM_WORLD.Get_size() - 1
rank = comm.Get_rank()
worker_ranks = set(range(1, MPI.COMM_WORLD.Get_size()))
rounds = 20
total_num_mess = rounds*num_workers

array_size = int(1e6)   # Size of large array in sim_specs

sim_specs = {'out': [('arr_vals', float, array_size),
                     ('scal_val', float)]}  # Test if get error without this

start_time = time.time()

if rank == 0:
    print("Running comms test on {} processors with {} workers".format(MPI.COMM_WORLD.Get_size(), num_workers))
    # print("Hello from manager")
    status = MPI.Status()
    alldone = False
    mess_count = 0
    while not alldone:
        for w in worker_ranks:
            if comm.Iprobe(source=w, tag=MPI.ANY_TAG, status=status):
                D_recv = comm.recv(source=w, tag=MPI.ANY_TAG, status=status)
                mess_count += 1
                # print('Message count', mess_count)
                # print('Message size (bytes): ', status.Get_count())

                # To test values
                x = w * 1000.0
                assert np.all(D_recv['arr_vals'] == x), "Array values do not all match"
                assert D_recv['scal_val'] == x + x/1e7, "Scalar values do not all match"
        if mess_count >= total_num_mess:
            alldone = True

    print('Manager received and checked {} messages'.format(mess_count))
    print('Manager finished in time', time.time() - start_time)

else:
    # print("Hello from worker", rank)
    output = np.zeros(1, dtype=sim_specs['out'])
    for x in range(rounds):
        x = rank * 1000.0
        output.fill(x)
        output['scal_val'] = x + x/1e7
        comm.send(obj=output, dest=0)



"""
Runs libEnsemble 1D sampling test with worker profiling.

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_1d_sampling_with_profile.py
   python3 test_1d_sampling_with_profile.py --nworkers 3 --comms local
   python3 test_1d_sampling_with_profile.py --nworkers 3 --comms tcp

The number of concurrent evaluations of the objective function will be 4-1=3.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: mpi local tcp
# TESTSUITE_NPROCS: 2 4
# TESTSUITE_EXTRA: true

import numpy as np
import os
import time

from libensemble.libE import libE
from libensemble.sim_funcs.one_d_func import one_d_example as sim_f
from libensemble.gen_funcs.sampling import latin_hypercube_sample as gen_f
from libensemble.tools import parse_args, add_unique_random_streams

nworkers, is_manager, libE_specs, _ = parse_args()

libE_specs['profile'] = True

sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
}

gen_specs = {
    'gen_f': gen_f,
    'out': [('x', float, (1,))],
    'user': {
        'gen_batch_size': 500,
        'lb': np.array([-3]),
        'ub': np.array([3]),
    },
}

persis_info = add_unique_random_streams({}, nworkers + 1)

exit_criteria = {'gen_max': 501}

# Perform the run
H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, libE_specs=libE_specs)

if is_manager:
    assert len(H) >= 501
    print("\nlibEnsemble with random sampling has generated enough points")

    assert 'manager.prof' in os.listdir(), 'Expected manager profile not found after run'
    os.remove('manager.prof')

    prof_files = ['worker_{}.prof'.format(i + 1) for i in range(nworkers)]

    # Ensure profile writes complete before checking
    time.sleep(0.5)

    for file in prof_files:
        assert file in os.listdir(), 'Expected profile {} not found after run'.format(file)
        with open(file, 'r') as f:
            data = f.read().split()
            num_worker_funcs_profiled = sum(['worker' in i for i in data])
        assert num_worker_funcs_profiled >= 8, (
            'Insufficient number of ' + 'worker functions profiled: ' + str(num_worker_funcs_profiled)
        )

        os.remove(file)



import os
from libensemble.resources.env_resources import EnvResources
from libensemble.resources import node_resources


def setup_standalone_run():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = ""


def teardown_standalone_run():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = ""


def setup_function(function):
    print("setup_function    function:%s" % function.__name__)
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = ""


def teardown_function(function):
    print("teardown_function    function:%s" % function.__name__)
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = ""

# Tests ========================================================================================


def test_get_cpu_resources_from_env_empty():
    # Test empty call
    cores_info = node_resources._get_cpu_resources_from_env()
    assert cores_info is None, 'cores_info should be None'


def test_get_cpu_resources_from_env_lsf():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = 'batch5' + ' g06n02' * 42
    exp_out = (42, 42)

    env_resources1 = EnvResources(nodelist_env_slurm="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                  nodelist_env_cobalt="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                  nodelist_env_lsf="LIBE_RESOURCES_TEST_NODE_LIST",
                                  nodelist_env_lsf_shortform="THIS_ENV_VARIABLE_IS_DEF_NOT_SET")
    cores_info = node_resources._get_cpu_resources_from_env(env_resources=env_resources1)
    assert cores_info == exp_out, "cores_info returned does not match expected"

    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = 'batch5' + ' g06n02' * 42 + ' h21n18' * 42
    env_resources2 = EnvResources(nodelist_env_slurm="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                  nodelist_env_cobalt="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                  nodelist_env_lsf="LIBE_RESOURCES_TEST_NODE_LIST",
                                  nodelist_env_lsf_shortform="THIS_ENV_VARIABLE_IS_DEF_NOT_SET")
    cores_info = node_resources._get_cpu_resources_from_env(env_resources=env_resources2)
    assert cores_info == exp_out, "cores_info returned does not match expected"


def test_get_cpu_resources_from_env_lsf_shortform():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = 'batch5 1 g06n02 42'
    exp_out = (42, 42)

    env_resources1 = EnvResources(nodelist_env_slurm="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                  nodelist_env_cobalt="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                  nodelist_env_lsf="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                  nodelist_env_lsf_shortform="LIBE_RESOURCES_TEST_NODE_LIST")
    cores_info = node_resources._get_cpu_resources_from_env(env_resources=env_resources1)
    assert cores_info == exp_out, "cores_info returned does not match expected"

    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = 'batch5 1 g06n02 42 h21n18 42'
    env_resources2 = EnvResources(nodelist_env_slurm="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                  nodelist_env_cobalt="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                  nodelist_env_lsf="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                  nodelist_env_lsf_shortform="LIBE_RESOURCES_TEST_NODE_LIST")
    cores_info = node_resources._get_cpu_resources_from_env(env_resources=env_resources2)
    assert cores_info == exp_out, "cores_info returned does not match expected"


def test_get_cpu_resources_from_env_unknown_env():
    os.environ["LIBE_RESOURCES_TEST_NODE_LIST"] = "knl-[0009-0012]"
    env_resources = EnvResources(nodelist_env_slurm="LIBE_RESOURCES_TEST_NODE_LIST",
                                 nodelist_env_cobalt="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_lsf="THIS_ENV_VARIABLE_IS_DEF_NOT_SET",
                                 nodelist_env_lsf_shortform="THIS_ENV_VARIABLE_IS_DEF_NOT_SET")

    cores_info = node_resources._get_cpu_resources_from_env(env_resources=env_resources)
    assert cores_info is None, 'cores_info should be None'


if __name__ == "__main__":
    setup_standalone_run()

    test_get_cpu_resources_from_env_empty()
    test_get_cpu_resources_from_env_lsf()
    test_get_cpu_resources_from_env_lsf_shortform()
    test_get_cpu_resources_from_env_unknown_env()

    teardown_standalone_run()



""" Sim function for Rosenbrock function. We include a @const input to the
    gradient, and not the function evaluation, to scale down the gradient
    so the Lipschitz and smoothness term is reduced.
"""
import numpy as np


def EvaluateFunction(x, component):
    """
    Evaluates the chained Rosenbrock function

    Parameters
    ----------
    x : np.ndarray
        - input vector
    component : int
        - index
    """

    assert len(x) % 2 == 0, print("must be even lengthed input vector")

    n = len(x) // 2

    if np.isnan(component):
        f1 = 100 * np.power(np.power(x[::2], 2) - x[1::2], 2)
        f2 = np.power(x[::2]-np.ones(n), 2)
        f = f1 + f2
    else:
        i = component
        x1 = x[2*i]
        x2 = x[2*i+1]
        f = 100 * (x1**2 - x2)**2 + (x1-1)**2

    return f


def EvaluateJacobian(x, component, const):
    """
    Evaluates the chained Rosenbrock Jacobian

    Parameters
    ----------
    x : np.ndarray
        - input vector
    component : int
        - index
    const : float
        - term to scale gradient by
    """

    assert len(x) % 2 == 0, print("must be even lengthed input vector")

    n = len(x) // 2
    df = np.zeros(len(x), dtype=float)

    if np.isnan(component):
        df[::2] = 400 * np.multiply(x[::2], np.power(x[::2], 2) - x[1::2]) + 2 * (x[::2] - np.ones(n))
        df[1::2] = -200 * (np.power(x[::2], 2) - x[1::2])

    else:
        i = component
        x1 = x[2*i]
        x2 = x[2*i+1]

        df[2*i] = 400 * x1 * (x1**2 - x2) + 2 * (x1 - 1)
        df[2*i+1] = -200 * (x1**2 - x2)

    return 1.0/const * df


def rosenbrock_eval(H, persis_info, sim_specs, _):

    if 'params' in persis_info:
        const = persis_info['params'].get('const', 1000)
    else:
        const = 1000

    batch = len(H['x'])
    H_o = np.zeros(batch, dtype=sim_specs['out'])

    for i, x in enumerate(H['x']):
        if 'obj_component' in H.dtype.fields:
            obj_component = H['obj_component'][i]  # which f_i

            if H[i]['get_grad']:
                H_o['gradf_i'][i] = EvaluateJacobian(x, obj_component, const)
            else:
                H_o['f_i'][i] = EvaluateFunction(x, obj_component)

        else:

            if persis_info.get('get_grad', False):
                H_o['grad'][i] = EvaluateJacobian(x, np.nan, const)

            H_o['f'][i] = EvaluateFunction(x, np.nan)
    return H_o, persis_info



"""
Tests the APOSMM generator function's ability to handle exceptions

Execute via one of the following commands (e.g. 3 workers):
   mpiexec -np 4 python3 test_persistent_aposmm_exception.py
   python3 test_persistent_aposmm_exception.py --nworkers 3 --comms local

When running with the above commands, the number of concurrent evaluations of
the objective function will be 2, as one of the three workers will be the
persistent generator.
"""

# Do not change these lines - they are parsed by run-tests.sh
# TESTSUITE_COMMS: local mpi
# TESTSUITE_NPROCS: 4
# TESTSUITE_EXTRA: true

import sys
import numpy as np

# Import libEnsemble items for this test
from libensemble.libE import libE
from libensemble.sim_funcs.periodic_func import func_wrapper as sim_f

import libensemble.gen_funcs

libensemble.gen_funcs.rc.aposmm_optimizers = 'nlopt'
from libensemble.gen_funcs.persistent_aposmm import aposmm as gen_f

from libensemble.alloc_funcs.persistent_aposmm_alloc import persistent_aposmm_alloc as alloc_f
from libensemble.tools import parse_args, add_unique_random_streams


def assertion(passed):
    """Raise assertion or MPI Abort"""
    if libE_specs['comms'] == 'mpi':
        from mpi4py import MPI

        if passed:
            print("\n\nMPI will be aborted as planned\n\n", flush=True)
            MPI.COMM_WORLD.Abort(0)  # Abort with success
        else:
            MPI.COMM_WORLD.Abort(1)  # Abort with failure
    else:
        assert passed
        print("\n\nException received as expected")


nworkers, is_manager, libE_specs, _ = parse_args()

if nworkers < 2:
    sys.exit("Cannot run with a persistent worker if only one worker -- aborting...")

n = 2
sim_specs = {
    'sim_f': sim_f,
    'in': ['x'],
    'out': [('f', float)],
}

gen_out = [('x', float, n), ('x_on_cube', float, n), ('sim_id', int), ('local_min', bool), ('local_pt', bool)]

gen_specs = {
    'gen_f': gen_f,
    'persis_in': ['f'] + [n[0] for n in gen_out],
    'out': gen_out,
    'user': {
        'initial_sample_size': 100,
        'localopt_method': 'LN_BOBYQA',
        'lb': np.array([0, -np.pi / 2]),
        'ub': np.array([2 * np.pi, 3 * np.pi / 2]),
    },
}

alloc_specs = {'alloc_f': alloc_f}

exit_criteria = {'sim_max': 1000}

persis_info = add_unique_random_streams({}, nworkers + 1)

libE_specs['abort_on_exception'] = False
try:
    # Perform the run, which will fail because we want to test exception handling
    H, persis_info, flag = libE(sim_specs, gen_specs, exit_criteria, persis_info, alloc_specs, libE_specs)
except Exception as e:
    if is_manager:
        if e.args[1].endswith('NLopt roundoff-limited'):
            assertion(True)
        else:
            assertion(False)
else:
    if is_manager:
        assertion(False)



"""
@Author: Created by Caleb Ju at Argonne National Labs as Given's associate,
         Summer 2021.
@About: Based on communication sliding primal-dual method:
        https://link.springer.com/article/10.1007/s10107-018-1355-4
"""
import numpy as np
from libensemble.message_numbers import STOP_TAG, PERSIS_STOP, FINISHED_PERSISTENT_GEN_TAG
from libensemble.tools.consensus_subroutines import (print_final_score, get_grad,
                                                     get_consensus_gradient, get_grad_locally)


def opt_slide(H, persis_info, gen_specs, libE_info):
    """ Gradient sliding. Coordinates with alloc to do local and distributed
        (i.e., gradient of consensus step) calculations.
    """
    # Send batches until manager sends stop tag
    tag = None
    local_gen_id = persis_info['worker_num']

    # each gen has unique internal id
    ub = gen_specs['user']['ub']
    lb = gen_specs['user']['lb']
    n = len(ub)

    f_i_idxs = persis_info['f_i_idxs']

    # start with random x_0
    x_0 = persis_info['rand_stream'].uniform(low=lb, high=ub, size=(n,))
    # ===== NOTATION =====
    # x_hk == \hat{x}_k
    # x_tk == \tilde{x}_k
    # x_uk == \underline{x}_k
    # prev_x_k == x_{k-1}
    # prevprev_x_k = x_{k-2}
    # ====================
    prev_x_k = x_0.copy()
    prev_x_uk = x_0.copy()
    prev_x_hk = x_0.copy()
    prev_z_k = np.zeros(len(x_0), dtype=float)
    prevprev_x_k = x_0.copy()
    prev_penult_k = x_0.copy()

    mu = persis_info['params']['mu']
    L = persis_info['params']['L']
    A_norm = persis_info['params']['A_norm']
    Vx_0x = persis_info['params']['Vx_0x']
    eps = persis_info['params']['eps']
    f_i_eval = persis_info['params'].get('f_i_eval', None)
    df_i_eval = persis_info['params'].get('df_i_eval', None)

    R = 1.0/(4 * (Vx_0x)**0.5)
    N = int(4 * (L*Vx_0x/eps)**0.5 + 1)

    weighted_x_hk_sum = np.zeros(len(x_0), dtype=float)
    b_k_sum = 0

    prev_b_k = 0
    prev_T_k = 0

    if local_gen_id == 1:
        print('[{}%]: '.format(0), flush=True, end='')
    print_final_score(prev_x_k, f_i_idxs, gen_specs, libE_info)
    percent = 0.1

    for k in range(1, N+1):
        tau_k = (k-1)/2
        lam_k = (k-1)/k
        b_k = k
        p_k = 2*L/k
        T_k = int(k*R*A_norm/L + 1)

        x_tk = prev_x_k + lam_k*(prev_x_hk - prevprev_x_k)
        x_uk = (x_tk + tau_k*prev_x_uk)/(1+tau_k)

        if f_i_eval is not None:
            y_k = get_grad_locally(x_uk, f_i_idxs, df_i_eval)
        else:
            tag, y_k = get_grad(x_uk, f_i_idxs, gen_specs, libE_info)

        if tag in [STOP_TAG, PERSIS_STOP]:
            return None, persis_info, FINISHED_PERSISTENT_GEN_TAG

        settings = {'T_k': T_k,
                    'b_k': k,
                    'p_k': p_k,
                    'mu': mu,
                    'L': L,
                    'R': R,
                    'k': k,
                    'prev_b_k': prev_b_k,
                    'prev_T_k': prev_T_k,
                    }

        [tag, x_k, x_k_1, z_k, x_hk] = primaldual_slide(y_k, prev_x_k,
                                                        prev_penult_k, prev_z_k,
                                                        settings, gen_specs,
                                                        libE_info)
        if tag in [STOP_TAG, PERSIS_STOP]:
            return None, persis_info, FINISHED_PERSISTENT_GEN_TAG

        prevprev_x_k = prev_x_k
        prev_x_k = x_k
        prev_x_hk = x_hk
        prev_penult_k = x_k_1  # penultimate x_k^{(i)}
        prev_z_k = z_k
        prev_b_k = b_k
        prev_T_k = T_k
        prev_x_uk = x_uk

        weighted_x_hk_sum += b_k * x_hk
        b_k_sum += b_k

        if k/N >= percent:
            curr_x_star = 1.0/b_k_sum * weighted_x_hk_sum
            if local_gen_id == 1:
                print('[{}%]: '.format(int(percent*100)), flush=True, end='')
            percent += 0.1
            print_final_score(curr_x_star, f_i_idxs, gen_specs, libE_info)

    return None, persis_info, FINISHED_PERSISTENT_GEN_TAG


def primaldual_slide(y_k, x_curr, x_prev, z_t, settings, gen_specs, libE_info):
    tag = None

    # define params
    T_k = settings['T_k']
    b_k = settings['b_k']
    p_k = settings['p_k']
    mu = settings['mu']
    L = settings['L']
    R = settings['R']
    k = settings['k']
    prev_b_k = settings['prev_b_k']
    prev_T_k = settings['prev_T_k']

    x_k_1 = x_curr.copy()
    xsum = np.zeros(len(x_curr), dtype=float)

    for t in range(1, T_k+1):
        # define per-iter params
        eta_t = (p_k + mu)*(t-1) + p_k*T_k
        q_t = L*T_k/(2*b_k*R**2)
        if k >= 2 and t == 1:
            a_t = prev_b_k*T_k/(b_k*prev_T_k)
        else:
            a_t = 1

        u_t = x_curr + a_t * (x_curr - x_prev)

        # compute first argmin
        tag, Lu_t = get_consensus_gradient(u_t, gen_specs, libE_info)
        if tag in [STOP_TAG, PERSIS_STOP]:
            return [tag, None, None, None, None]

        z_t = z_t + (1.0/q_t) * Lu_t

        # computes second argmin
        tag, Lz_t = get_consensus_gradient(z_t, gen_specs, libE_info)
        if tag in [STOP_TAG, PERSIS_STOP]:
            return [tag, None, None, None, None]

        x_next = (eta_t*x_curr) + (p_k*x_k_1) - (y_k + Lz_t)
        x_next /= (eta_t + p_k)

        x_prev = x_curr
        x_curr = x_next

        xsum += x_curr
        # zsum += z_t

    x_k = x_curr
    x_k_1 = x_prev
    z_k = z_t
    x_hk = xsum/T_k

    return [tag, x_k, x_k_1, z_k, x_hk]
